# this file is generated by Lino
from django.utils.translation import ugettext
ugettext('Next page')
ugettext('Please wait...')
ugettext('Action not available on phantom record.')
ugettext('Next')
ugettext('Go to record')
ugettext('Reload current record')
ugettext('Save changes to current record ?')
ugettext('Edit text in own window')
ugettext("Field '{0}' is disabled")
ugettext('Confirmation')
ugettext('No data to display.')
ugettext('Alert')
ugettext('Previous page')
ugettext('Cannot show detail for empty foreign key.')
ugettext('This record is disabled')
ugettext('View')
ugettext('Displaying {0} - {1} of {2}')
ugettext('Team view')
ugettext('Last')
ugettext('Edit help texts for fields on this model.')
ugettext('Last page')
ugettext('of {0}')
ugettext('Help Text Editor')
ugettext('Save Grid Configuration')
ugettext('Page')
ugettext('First')
ugettext('Show this panel in own window')
ugettext('[parameters]')
ugettext('This field is disabled')
ugettext('Layout Editor')
ugettext('Save changes to text ?')
ugettext('No more records to display. Detail window has been closed.')
ugettext('First page')
ugettext('Select another view of this report')
ugettext('Cannot show MTI child if checkbox is off.')
ugettext('Previous')
ugettext('Edit Detail layout')
ugettext('Enter a date in format YYYY-MM-DD (use zeroes for unknown parts).')



# -*- coding: UTF-8 -*-
## Copyright 2012 Luc Saffre
## This file is part of the Lino project.
## Lino is free software; you can redistribute it and/or modify 
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 3 of the License, or
## (at your option) any later version.
## Lino is distributed in the hope that it will be useful, 
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
## GNU General Public License for more details.
## You should have received a copy of the GNU General Public License
## along with Lino; if not, see <http://www.gnu.org/licenses/>.

import decimal
from dateutil.relativedelta import relativedelta
ONE_DAY = relativedelta(days=1)

from django.db import models
from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.utils.translation import ugettext as _


from lino.utils import i2d
from lino.utils.instantiator import Instantiator
from lino.core.modeltools import resolve_model
from lino.utils.babel import babel_values

from lino.modlib.debts.models import AccountType

def n2dec(v):
    return decimal.Decimal("%.2d" % v)
   
def objects():
    
    User = resolve_model('users.User')
    kerstin = User(username="kerstin",
        first_name="Kerstin",last_name=u"Küpper",
        profile='300') # UserProfiles.kerstin)
        #~ level=UserLevel.user,
        #~ debts_level=UserLevel.user)
    yield kerstin
    
    
    Household = resolve_model('households.Household')
    Budget = resolve_model('debts.Budget')
    Actor = resolve_model('debts.Actor')
    for hh in Household.objects.all():
        b = Budget(partner_id=hh.id,user=kerstin)
        b.fill_defaults(None)
        yield b
        
    Budget = resolve_model('debts.Budget')
    Entry = resolve_model('debts.Entry')
    Account = resolve_model('debts.Account')
    Company = resolve_model('contacts.Company')
    INCOME_AMOUNTS = Cycler([i*200 for i in range(8)])
    EXPENSE_AMOUNTS = Cycler([i*5.24 for i in range(10)])
    DEBT_AMOUNTS = Cycler([(i+1)*300 for i in range(5)])
    PARTNERS = Cycler(Company.objects.all())
    LIABILITIES = Cycler(Account.objects.filter(type=AccountType.liability))
    for b in Budget.objects.all():
        #~ n = min(3,b.actor_set.count())
        for e in b.entry_set.all():
            #~ for i in range(n):
            if e.account.type == AccountType.income:
                amount = INCOME_AMOUNTS.pop()
            elif e.account.type == AccountType.expense:
                amount = EXPENSE_AMOUNTS.pop()
            if e.account.required_for_household:
                e.amount = n2dec(amount)
            if e.account.required_for_person:
                for a in b.actor_set.all():
                    e.amount = n2dec(amount)
                    e.actor = a
            e.save()
        ACTORS = Cycler(None,*[a for a in b.actor_set.all()])
        for i in range(4):
            amount = int(DEBT_AMOUNTS.pop())
            kw = dict(budget=b,
                account=LIABILITIES.pop(),
                partner=PARTNERS.pop(),
                amount=amount,
                actor=ACTORS.pop())
            if amount > 600:
                kw.update(distribute=True)
            else:
                kw.update(monthly_rate=n2dec(amount/20))
            yield Entry(**kw)
    
    


import cgi
import datetime
from django.utils import timezone
from django.db import models
from django.conf import settings

from lino import dd
from lino.utils import babel


class Poll(dd.Model):
    question = models.CharField("Question text", max_length=200)
    hidden = models.BooleanField("Hidden",help_text="""\
Whether this poll should not be shown in the main window.""")
    pub_date = models.DateTimeField('Date published',auto_now_add=True)
    
    class Meta:
        verbose_name = 'Poll'
        verbose_name_plural = 'Polls'
    
    def __unicode__(self):
        return self.question
        
    def was_published_recently(self):
        return self.pub_date >= timezone.now() - datetime.timedelta(days=1)        
        
class Choice(dd.Model):
    poll = models.ForeignKey(Poll)
    choice = models.CharField("Choice text",max_length=200)
    votes = models.IntegerField("No. of votes",default=0)
    
    class Meta:
        verbose_name = 'Choice'
        verbose_name_plural = 'Choices'
    
    def __unicode__(self):
        return self.choice    

    @dd.action(help_text="Click here to see how Lino implements custom actions.")
    def vote(self,ar,**kw):
        if self.votes > 2:
            msg = "%s has already %d votes!" % (self,self.votes)
            msg += "\nDo you still want to vote for it?"
            ar.confirm(msg)
        self.votes += 1
        self.save()
        kw.update(refresh=True)
        kw.update(alert="Voted!")
        kw.update(message="Thank you for voting %s" % self)
        return ar.success_response(**kw)
        


class Polls(dd.Table):
    model = Poll
    sort_order = ['pub_date']
    
    detail_layout = """
    id question 
    hidden pub_date
    ChoicesByPoll
    """
    
    insert_layout = dd.FormLayout("""
    question
    hidden
    """,window_size=(40,'auto'))
    
    

class Choices(dd.Table):
    model = Choice
        
class ChoicesByPoll(Choices):
    master_key = 'poll'
    

def recent_polls(request):
    html = '<h1>%s</h1> ' % cgi.escape("Recent polls")
    html += '<ul>'
    for poll in Poll.objects.filter(hidden=False).order_by('pub_date'):
        html += '<li>'
        html += '<b>%s</b> ' % cgi.escape(poll.question)
        chunks = []
        for obj in poll.choice_set.all():
            chunks.append(settings.LINO.ui.row_action_button(obj,request,Choices.vote,unicode(obj)))
        html += ' / '.join(chunks)
        
        html += "<br/><small>Published %s" % babel.dtosl(poll.pub_date)
        chunks = []
        for obj in poll.choice_set.all():
            chunks.append("%d %s" % (obj.votes,cgi.escape(unicode(obj))))
        html += '<br/>Results: %s' % (', '.join(chunks))
        html += '</small>'
        html += '</li>'
    html += '</ul>'
    return html
    



# -*- coding: UTF-8 -*-
## Copyright 2011-2012 Luc Saffre
## This file is part of the Lino project.
## Lino is free software; you can redistribute it and/or modify 
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 3 of the License, or
## (at your option) any later version.
## Lino is distributed in the hope that it will be useful, 
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
## GNU General Public License for more details.
## You should have received a copy of the GNU General Public License
## along with Lino; if not, see <http://www.gnu.org/licenses/>.

import datetime
import traceback
from django.conf import settings
from django.db import models
from django.utils.translation import ugettext_lazy as _
from django.utils.encoding import force_unicode


from lino import mixins
from lino import dd
from lino.utils import Warning
from lino.utils import join_words
from lino.utils import AttrDict, IncompleteDate
from lino.core.modeltools import obj2str

from lino.utils import babel

from lino.utils.xmlgen import html as xghtml
E = xghtml.E

from lino.utils.babel import dtos
from lino.utils.choicelists import ChoiceList



try:
    import suds
except ImportError, e:
    pass


from lino.modlib.cbss.models import NewStyleRequest,SSIN, get_client, \
  CBSSRequestDetail, CBSSRequests, cbss2gender, RequestStates, CBSS_ERROR_MESSAGE
  
class RequestLanguages(ChoiceList):
    label = _("Language")
add = RequestLanguages.add_item
add("nl",_("Dutch"),"nl")
add("fr",_("French"),"fr")
add("de",_("German"),"de")

class RetrieveTIGroupsRequest(NewStyleRequest,SSIN):
    """
    A request to the RetrieveTIGroups service (aka Tx25)
    """
    
    class Meta:
        verbose_name = _("Tx25 Request")
        verbose_name_plural = _('Tx25 Requests')
        
    wsdl_parts = ('cache','wsdl','RetrieveTIGroupsV3.wsdl')
    
    #~ language = babel.LanguageField()
    language = RequestLanguages.field()
    history = models.BooleanField(
        verbose_name=_("History"),default=False,
        help_text = "Whatever this means.")
        
    def get_print_language(self,pm):
        if self.language.value in babel.AVAILABLE_LANGUAGES:
            return self.language.value
        return babel.DEFAULT_LANGUAGE
        
    def fill_from_person(self,person):
        self.national_id = person.national_id
        self.language = person.language # .value # babel.DEFAULT_LANGUAGE
        
        
    def get_service_reply(self,**kwargs):
        client = get_client(self)
        meth = client.service.retrieveTI
        clientclass = meth.clientclass(kwargs)
        client = clientclass(meth.client, meth.method)
        #~ print 20120613, portSelector[0]
        #~ print '20120613b', dir(client)
        return client.succeeded(client.method.binding.input, self.response_xml.encode('utf-8'))
        
    
    def execute_newstyle(self,client,infoCustomer,simulate_response):
        si = client.factory.create('ns0:SearchInformationType')
        si.ssin = self.get_ssin()
        if self.language:
            si.language = self.language.value
        #~ if self.history:
            #~ si.history = 'true'
        si.history = self.history
        #~ if validate:
            #~ self.validate_newstyle(srvreq)
        if simulate_response is None:
            self.check_environment(si)
            try:
                reply = client.service.retrieveTI(infoCustomer,None,si)
            except suds.WebFault,e:
                """
                Example of a SOAP fault:
          <soapenv:Fault>
             <faultcode>soapenv:Server</faultcode>
             <faultstring>An error occurred while servicing your request.</faultstring>
             <detail>
                <v1:retrieveTIGroupsFault>
                   <informationCustomer xmlns:ns0="http://kszbcss.fgov.be/intf/RetrieveTIGroupsService/v1" xmlns:ns1="http://schemas.xmlsoap.org/soap/envelope/">
                      <ticket>2</ticket>
                      <timestampSent>2012-05-23T10:19:27.636628+01:00</timestampSent>
                      <customerIdentification>
                         <cbeNumber>0212344876</cbeNumber>
                      </customerIdentification>
                   </informationCustomer>
                   <informationCBSS>
                      <ticketCBSS>f4b9cabe-e457-4f6b-bfcc-00fe258a9b7f</ticketCBSS>
                      <timestampReceive>2012-05-23T08:19:09.029Z</timestampReceive>
                      <timestampReply>2012-05-23T08:19:09.325Z</timestampReply>
                   </informationCBSS>
                   <error>
                      <severity>FATAL</severity>
                      <reasonCode>MSG00003</reasonCode>
                      <diagnostic>Unexpected internal error occurred</diagnostic>
                      <authorCode>http://www.bcss.fgov.be/en/international/home/index.html</authorCode>
                   </error>
                </v1:retrieveTIGroupsFault>
             </detail>
          </soapenv:Fault>
                """
                
                msg = CBSS_ERROR_MESSAGE % e.fault.faultstring
                msg += unicode(e.document)
                self.status = RequestStates.failed
                raise Warning(msg)
            self.response_xml = reply
        else:
            self.response_xml = simulate_response
            
        #~ self.response_xml = unicode(reply)
        reply = self.get_service_reply()
        self.ticket = reply.informationCBSS.ticketCBSS
        if reply.status.value == "NO_RESULT":
            msg = CBSS_ERROR_MESSAGE % reply.status.code
            keys = ('value','code','description')
            msg += '\n'.join([
                k+' : '+getattr(reply.status,k)
                    for k in keys])
            for i in reply.status.information:
                msg += "\n- %s = %s" % (i.fieldName,i.fieldValue)
            self.status = RequestStates.warnings
            raise Warning(msg)
            
        self.status = RequestStates.ok
        #~ self.response_xml = str(res)
        #~ self.response_xml = "20120522 %s %s" % (res.__class__,res)
        #~ print 20120523, res.informationCustomer
        #~ print self.response_xml
        return reply
        
          
    def Result(self,ar):
        return ar.spawn(RetrieveTIGroupsResult,master_instance=self)
        
        
  
class RetrieveTIGroupsRequestDetail(CBSSRequestDetail):
  
    parameters = "national_id language history"
    
    result = "cbss.RetrieveTIGroupsResult"
    
    #~ def setup_handle(self,lh):
        #~ CBSSRequestDetail.setup_handle(self,lh)
        
class RetrieveTIGroupsRequestInsert(dd.FormLayout):
    window_size = (40,'auto')
    main = """
    person
    national_id language 
    history
    """

class RetrieveTIGroupsRequests(CBSSRequests):
    #~ debug_permissions = True
    model = RetrieveTIGroupsRequest
    detail_layout = RetrieveTIGroupsRequestDetail()
    insert_layout = RetrieveTIGroupsRequestInsert()
    #~ insert_layout = RetrieveTIGroupsRequestInsert(window_size=(400,'auto'))
    required_user_groups = ['cbss']
        
    #~ @dd.virtualfield(dd.HtmlBox())
    #~ def result(self,row,ar):
        #~ return row.response_xml
        
class RetrieveTIGroupsRequestsByPerson(RetrieveTIGroupsRequests):
    master_key = 'person'
    
class MyRetrieveTIGroupsRequests(RetrieveTIGroupsRequests,mixins.ByUser):
    pass
    





def rn2date(rd):
    return IncompleteDate(int(rd.Century+rd.Year),int(rd.Month),int(rd.Day))
    
def deldate(n):
    if hasattr(n,'DelDate'):
        return [' (' + unicode(_('until ')) + dtos(rn2date(n.DelDate)) + ')']
    #~ return [' (' + unicode(_('until today')) + ')']
    return []
    #~ return []
    
    
#~ def simpleattr(n,name):
    #~ v = getattr(n,name,None)
    #~ if v:
        #~ return [ ', '+name+' ' + unicode(v)]
    #~ return []
    
def simpletype(v):
    return Info(xghtml.E.b(unicode(v)))
def boldstring(v):
    return Info(xghtml.E.b(unicode(v)))
    
class Info(object):
    def __init__(self,*chunks):
        self.chunks = list(chunks)
    def addfrom(self,node,name,prefix=None,fmt=boldstring,suffix=''):
        v = getattr(node,name,None)
        if not v: return self
        if prefix is None:
            prefix = '%s ' % name
        else:
            prefix = force_unicode(prefix)
        if len(self.chunks):
            if not prefix.startswith(' '):
                prefix = ', ' + prefix
        if prefix and prefix[-1] not in ' :':
            prefix += ': '
        self.chunks += [prefix] + fmt(v).chunks
        if suffix:
            self.chunks.append(force_unicode(suffix))
        return self
    def add_deldate(self,n):
        self.chunks += deldate(n)
    def add_codelabel(self,n):
        self.chunks += code_label(n).chunks
        #~ if hasattr(n,'Label'):
            #~ self.addfrom(n,'Label')
            #~ self.addfrom(n,'Code','(',simpletype,')')
        #~ else:
            #~ self.addfrom(n,'Code','[',boldstring,']')
        return self
    
def code_label(n):
    if hasattr(n,'Label') and n.Label:
        return Info(E.b(n.Label),' (',n.Code,')')
    return Info(E.b(n.Code))

def datarow(group,node,since,info):
    if group and node.__class__.__name__.startswith('IT'):
        itnum = node.__class__.__name__[2:]
    else:
        itnum = ''
    if hasattr(node,'Type'):
        group += " " + node.Type
        group += " " + node.Status
        group += " " + node.Structure
    return AttrDict(group=group,
        type=itnum,since=rn2date(since),info=E.p(*info.chunks))
    
#~ CodeLabel = code_label
#~ def CodeLabel(n):
    #~ info = Info()
    #~ return info
    
def NameType(n):
    info = Info()
    s = ' '.join([ln.Label for ln in n.LastName])
    info.chunks.append(E.b(s))
    info.chunks.append(', ')
    s = ' '.join([fn.Label for fn in n.FirstName])
    info.chunks.append(s)
    return info
    
    

#~ def addinfo(node,name,prefix=None,fmt=simpletype,suffix=''):
    #~ v = getattr(node,name,None)
    #~ if not v: return []
    #~ if prefix is None:
        #~ prefix = ', %s ' % name
    #~ info = [force_unicode(prefix)] + fmt(v)
    #~ if suffix:
        #~ info.append(force_unicode(suffix))
    #~ return info
    
def DateType(n):
    return Info(babel.dtos(rn2date(n)))
    
def ForfeitureDateType(n):
    info = Info(babel.dtos(rn2date(n)))
    info.addfrom(n,'Graphic',' (',simpletype,')')
    return info
    
def ExpiryDateType(n):
    info = Info(babel.dtos(rn2date(n)))
    info.addfrom(n,'Graphic',' (',simpletype,')')
    return info
    
def TribunalType(n):
    return code_label(n)
def PlaceType(n):
    return code_label(n)
def GraphicPlaceType(n):
    info = CountryType(n.Country)
    info.addfrom(n,'Graphic','')
    #~ if hasattr(n,'Graphic'):
        #~ info.append(', graphic:'+n.Graphic)
    return info
def ForeignJudgementType(n):
    return GraphicPlaceType(n.Place)
def BelgianJudgementType(n):
    info = Info()
    info.addfrom(n,'Tribunal',None,TribunalType)
    info.addfrom(n,'Date',None,DateType)
    info.addfrom(n,'Place',None,PlaceType)
    #~ info += TribunalType(n.Tribunal)
    #~ info += DateType(n.Date)
    #~ info += PlaceType(n.Place)
    return info
def CountryType(n):
    return code_label(n)
def LieuType(n):
    info = Info()
    if hasattr(n,'Place1'):
        #~ info += code_label(n.Place1)
        info.addfrom(n,'Place1',None,code_label)
    elif hasattr(n,'Place2'):
        info.addfrom(n,'Place2',None,GraphicPlaceType)
    else:
        place = n.Place3
        #~ info += GraphicPlaceType(place)
        info.addfrom(place,'BelgianJudgement','',BelgianJudgementType)
        info.addfrom(place,'ForeignJudgement','',ForeignJudgementType)
        #~ if hasattr(place,'BelgianJudgement'):
            #~ info += BelgianJudgementType(place.BelgianJudgement)
        #~ else:
            #~ info += ForeignJudgementType(place.ForeignJudgement)
    return info

def DeliveryType(n):
    return PlaceType(n.Place)
    
def DiplomaticPostType(n):
    return code_label(n)
def TerritoryType(n):
    return code_label(n)
def ProvinceType(n):
    return code_label(n)
    
def IssuerType(n):
    # prefixes can be empty since this is a xs:choice
    info = Info().addfrom(n,'Place','',PlaceType)
    info.addfrom(n,'Province','',ProvinceType,' (%s)' % unicode(_("Province")))
    info.addfrom(n,'PosteDiplomatique','',DiplomaticPostType,' (%s)' % unicode(_("Diplomatic post")))
    return info

def ResidenceType(n):
    return code_label(n)
    
def NationalNumberType(n):
    info = Info().addfrom(n,'NationalNumber','')
    return info # [n.NationalNumber]
    
def PartnerType(n):
    info = Info().addfrom(n,'NationalNumber','',NationalNumberType)
    #~ info.addfrom(n,'Name','',NameType)
    info.addfrom(n,'Name',' ',NameType)
    return info
    
def NotaryType(n):
    info = Info().addfrom(n,'NameNotary')
    info.addfrom(n,'Place',' in ',PlaceType)
    info.addfrom(n,'Country',', ',CountryType)
    return info
    
def NotificationType(n):
    info = Info().addfrom(n,'NotificationDate',None,DateType)
    info.addfrom(n,'Place',' in ',PlaceType)
    return info
    
def ReasonType(n):
    return code_label(n)
    
def CessationType(n):
    return code_label(n)
    
def DeclarationType(n):
    return code_label(n)
    
def Residence(n):
    info = Info().addfrom(n,'Residence','',ResidenceType)
    info.addfrom(n,'Fusion',_("Fusion"))
    info.addfrom(n,'Language',_("Language"))
    info.add_deldate(n)
    return info
    
    
def IT003(n): # AscertainedLegalMainAddresses : Détermination de résidence
    #~ raise Exception(str(n))
    def InvestigationResultType(n): 
        return code_label(n)
    info = Info().addfrom(n,'InvestigationResult','',InvestigationResultType)
    info.addfrom(n,'Graphic1','')
    info.addfrom(n,'Graphic2','')
    info.add_deldate(n)
    return info
  
def IT005(n): # AddressChangeIntention
    #~ raise Exception(str(n))
    info = Info().addfrom(n,'OriginPlace',_('Move from '),PlaceType)
    info.addfrom(n,'DestinationPlace',_('Move to '),PlaceType)
    info.add_deldate(n)
    return info
  
def IT006(n):
    info = Info()
    info.addfrom(n,'Country','',CountryType)
    info.addfrom(n,'Graphic',' ')
    info.add_deldate(n)
    return info

def IT018(n):
    info = Info()
    info.addfrom(n,'Address','',AddressType)
    info.add_deldate(n)
    return info

def IT024(n): 
    info = Info()
    info.add_deldate(n)
    return info
  
def FiliationType(n):
    return code_label(n)
def ParentType(n):
    info = Info()
    info.addfrom(n,'Name','',NameType)
    info.addfrom(n,'NationalNumber',' (',NationalNumberType,')')
    return info
def IT110(n):
    #~ info = code_label(n.FiliationType)
    #~ info.addfrom(n.Parent1,'Name','',NameType)
    #~ info.addfrom(n.Parent1,'NationalNumber',' (',NationalNumberType,')')
    #~ info.chunks.append(_('and '))
    #~ info.addfrom(n.Parent2,'Name','',NameType)
    #~ info.addfrom(n.Parent2,'NationalNumber',' (',NationalNumberType,')')
    info = Info()
    info.addfrom(n,'FiliationType','',FiliationType)
    #~ info.chunks.append()
    info.addfrom(n,'Parent1',_('of '),ParentType)
    info.addfrom(n,'Parent2',_('and '),ParentType)
    info.addfrom(n,'ActNumber',_("Act no. "))
    info.addfrom(n,'Place',_("in "),PlaceType)
    info.addfrom(n,'Graphic'," ")
    info.add_deldate(n)
    return info
    
def IT140(n):
    info = Info().addfrom(n,'Name',' ',NameType)
    info.addfrom(n,'NationalNumber',' (',NationalNumberType,')')
    #~ info += _(' as ')
    info.addfrom(n,'FamilyRole',_('as '),code_label)
    info.addfrom(n,'Housing',None,HousingType)
    info.add_deldate(n)
    return info
    
def TypeOfLicenseType(n):
    return code_label(n)
    
def TypeOfLicenseType194(n):
    return code_label(n)
    
def DeliveryType194(n):
    info = Info().addfrom(n,'Place',_('in '),PlaceType)
    info.addfrom(n,'Label','')
    info.addfrom(n,'Code',' (',simpletype,')')
    #~ info.add_codelabel(n)
    #~ info += code_label(n)
    return info
    
def CategoryType(n):
    return code_label(n)
    
def GearBoxType(n):
    return code_label(n)
    
def MedicalType(n):
    return code_label(n)
    
def LicenseCategoriesType(n):
    info = Info()
    #~ raise Exception(str(n))
    #~ for cat in n.Category:
        #~ info.addfrom(cat,'Category',' ',CategoryType)
    info.chunks.append('/'.join([cat.Label for cat in n.Category]))
    #~ info += code_label(n)
    return info
    
def ForfeitureReasonType(n):
    return code_label(n)
    
def IT191(n):
    #~ info = code_label(n.TypeOfLicense)
    info = Info().addfrom(n,'TypeOfLicense','',TypeOfLicenseType)
    info.addfrom(n,'LicenseNumber',_('no. '))
    info.addfrom(n,'Place',_('delivered in '),PlaceType)
    info.addfrom(n,'DeliveryCountry',' (',CountryType,')')
    info.addfrom(n,'ForfeitureReason',None,ForfeitureReasonType)
    info.addfrom(n,'ForfeitureDate',None,ForfeitureDateType)
    #~ info.append()
    #~ info.append(E.b(n.LicenseNumber))
    #~ info.append(', categories ' 
      #~ + ' '.join([cat.Label for cat in n.Categories.Category]))
    #~ info.append(_(' delivered in '))
    #~ info += code_label(n.Delivery.Place)
    info.add_deldate(n)
    return info
    
def IT194(n):
    info = Info().addfrom(n,'TypeOfLicense','',TypeOfLicenseType194)
    info.addfrom(n,'Categories',_('categories '),LicenseCategoriesType)
    info.addfrom(n,'LicenseNumber',_('no. '))
    info.addfrom(n,'Delivery',_('delivered '),DeliveryType194)
    info.addfrom(n,'GearBox',None,GearBoxType)
    info.addfrom(n,'Medical',None,MedicalType)
    info.addfrom(n,'ExpiryDate',_('expires '),ExpiryDateType)
    info.add_deldate(n)
    return info
    
def IT198(n):
    info = Info().addfrom(n,'PermitNumber',_('no. '))
    info.addfrom(n,'Categories',_('categories '),LicenseCategoriesType)
    info.addfrom(n,'LicenseNumber',_('no. '))
    info.addfrom(n,'Delivery',_('delivered '),DeliveryType194)
    info.addfrom(n,'GearBox',None,GearBoxType)
    info.addfrom(n,'Medical',None,MedicalType)
    info.addfrom(n,'ExpiryDate',_('expires '),ExpiryDateType)
    info.add_deldate(n)
    return info
    



def TypeOfPassportType(n):
    return code_label(n)
    
def PassportIdentType(n):
    info = Info()
    info.addfrom(n,'PassportType',_('type '),TypeOfPassportType)
    info.addfrom(n,'PassportNumber',_('no. '))
    return info
    
def IT199(n):
    info = Info()
    #~ info.chunks.append('Number ')
    #~ info.chunks.append(E.b(n.PassportIdent.PassportNumber))
    #~ info.append(', status ')
    info.addfrom(n,'Status',_("status"),code_label)
    info.addfrom(n,'PassportIdent','',PassportIdentType)
    info.addfrom(n,'Issuer',_('issued by '),IssuerType)
    info.addfrom(n,'RenewalNumber',_('renewal no. '),boldstring)
    info.addfrom(n,'SerialNumber',_('serial no. '),boldstring)
    info.addfrom(n,'SecondNumber',_('second no. '),boldstring)
    info.addfrom(n,'ReplacementOf',_('replacement of '),boldstring)
    info.addfrom(n,'AdditionTo',_('addition to '),boldstring)
    info.addfrom(n,'ProductionDate',_('produced '),DateType)
    info.addfrom(n,'ExpiryDate',_('expires '),DateType)
    #~ info.append(', type ')
    #~ info += code_label(n.PassportIdent.PassportType)
    #~ info.append(', expires ')
    #~ info.append(E.b(dtos(rn2date(n.ExpiryDate))))
    #~ info.append(', delivered by ')
    #~ info += code_label(n.Issuer.PosteDiplomatique)
    #~ info.append(_(' renewal no. '))
    #~ info.append(E.b(n.RenewalNumber))
    info.add_deldate(n)
    return info

def HousingType(n):
    return code_label(n)
    
def AddressType(n):
    info = Info()
    #~ pd = n.Address.Address
    info.addfrom(n,'Country','',CountryType)
    #~ info.append(', ')
    info.addfrom(n,'Graphic1','')
    info.addfrom(n,'Graphic2','')
    info.addfrom(n,'Graphic3','')
    #~ info.append(E.b(pd.Graphic1))
    #~ info.append(', ')
    #~ info.append(E.b(pd.Graphic2))
    #~ info.append(', ')
    #~ info.append(E.b(pd.Graphic3))
    #~ info.addfrom(pd,'Graphic3')
    return info
    

def CertificateType(n):
    return code_label(n)
    




def IT200(n):
    info = Info().addfrom(n,'PublicSecurityNumber',_('no. '))
    info.add_deldate(n)
    return info
    
def IT202(n):
    info = Info()
    info.addfrom(n,'Graphic1','')
    info.addfrom(n,'Graphic2','')
    info.addfrom(n,'Limosa','',LimosaType)
    info.add_deldate(n)
    return info
def LimosaType(n):
    info = Info()
    info.addfrom(n,'Reason1','',LimosaReasonType)
    info.addfrom(n,'Reason2','',LimosaReasonType)
    info.addfrom(n,'NationalNumber',_('SSIN '),NationalNumberType)
    return info
def LimosaReasonType(n):
    return code_label(n)
    
def IT205(n):
    info = code_label(n)
    info.add_deldate(n)
    return info
    
def OrganizationType(n): return code_label(n)
def GeneralInfoType(n):
    info = code_label(n)
    info.addfrom(n,'Organization',_("Organization"),OrganizationType)
    return info
    
def OrigineType(n): 
    return Info().add_codelabel(n)
def AppealType(n):  return code_label(n)
def StatusAppealType(n):  return code_label(n)
def ProcedureType(n):
    info = Info()
    info.addfrom(n,'Origine',None,OrigineType)
    info.addfrom(n,'Reference')
    info.addfrom(n,'Appeal',None,AppealType)
    info.addfrom(n,'OpenClose',None,StatusAppealType)
    info.addfrom(n,'NationalNumber',_('SSIN '),NationalNumberType)
    return info
    
def DecisionCancelledType(n):
    info = Info()
    info.addfrom(n,'Date',None,DateType)
    info.addfrom(n,'Reference')
    return info
def DelayLeaveGrantedType(n):
    info = Info()
    info.addfrom(n,'Date',None,DateType)
    return info
def StrikingOutType(n):
    info = Info()
    info.addfrom(n,'Reference')
    info.addfrom(n,'OpenClose',None,OpenCloseType)
    info.addfrom(n,'Status',None,StrikingStatusType)
    return info
def StrikingStatusType(n):  return code_label(n)
def TerritoryLeftType(n):  return code_label(n)
def OpenCloseType(n):  return code_label(n)

def ProtectionType(n):
    info = code_label(n)
    info.addfrom(n,'Reference')
    info.addfrom(n,'Term')
    return info
    
def AdviceFromCGVSType(n):
    info = code_label(n)
    info.addfrom(n,'Reference')
    return info
    
def ApplicationFiledType(n):
    info = code_label(n)
    info.addfrom(n,'Place',_("in "),PlaceType)
    return info
    
def DecisionType206(n):
    info = code_label(n)
    info.addfrom(n,'Reference',_("Reference"))
    info.addfrom(n,'OpenClose',_("Open/Close"),OpenCloseType)
    info.addfrom(n,'Comments')
    info.addfrom(n,'Term')
    return info
    
def NotificationByDVZType(n):
    info = Info()
    info.addfrom(n,'Place',_("in "),PlaceType)
    info.addfrom(n,'Reference')
    return info
    
def NotificationByOrgType(n):
    info = Info()
    info.addfrom(n,'Reference')
    return info
    
def AppealLodgedType(n):
    info = Info()
    info.addfrom(n,'Reference')
    return info
    
def IT206(n):
    def Status(n):
        info = Info()
        info.addfrom(n,'Status')
        return info
        
    info = Info()
    info.addfrom(n,'GeneralInfo','',GeneralInfoType)
    info.addfrom(n,'Procedure',_("Procedure"),ProcedureType)
    info.addfrom(n,'StrikingOut',None,StrikingOutType)
    info.addfrom(n,'DecisionCancelled',_("Decision cancelled"),DecisionCancelledType)
    info.addfrom(n,'Protection',_("Protection"),ProtectionType)
    info.addfrom(n,'DelayLeaveGranted',None,DelayLeaveGrantedType)
    info.addfrom(n,'Escape',_("Escape"),Status)
    info.addfrom(n,'UnrestrictedStay',None,Status)
    info.addfrom(n,'ApplicationRenounced',_("Application renounced"),Status)
    info.addfrom(n,'TerritoryLeft',_("Territory left"),TerritoryLeftType)
    info.addfrom(n,'AdviceFromCGVS',None,AdviceFromCGVSType)
    info.addfrom(n,'Decision',_("Decision"),DecisionType206)
    info.addfrom(n,'ApplicationFiled',_("Application filed"),ApplicationFiledType)
    info.addfrom(n,'NotificationByDVZ',None,NotificationByDVZType)
    info.addfrom(n,'NotificationByOrg',None,NotificationByOrgType)
    info.addfrom(n,'AppealLodged',None,AppealLodgedType)
    info.add_deldate(n)
    return info
    


def InitiativeType(n):
    return code_label(n)
    
def SocialWelfareType(n):
    info = Info()
    info.addfrom(n,'Place',_("in "),PlaceType)
    info.addfrom(n,'Initiative',None,InitiativeType)
    info.add_deldate(n)
    return info
    
def RefugeeCentreType(n):
    return code_label(n)
    
def IT207(n):
    info = Info()
    info.addfrom(n,'SocialWelfare',_("Social Welfare Centre"),SocialWelfareType)
    info.addfrom(n,'RefugeeCentre',_("Refugee Centre"),RefugeeCentreType)
    info.add_deldate(n)
    return info
    
def RegistrationRegisterType(n):
    return code_label(n)
def IT210(n):
    info = Info()
    info.addfrom(n,'RegistrationRegister',_("Registration register"),RegistrationRegisterType)
    info.add_deldate(n)
    return info
    



def IdentificationType(n):
    return code_label(n)
def IT211(n):
    info = Info()
    info.addfrom(n,'TypeOfDocument','',IdentificationType)
    info.add_deldate(n)
    return info
    




def ChoosenResidenceType(n):
    return code_label(n)
def IT212(n):
    info = Info().addfrom(n,'Residence',None,ChoosenResidenceType)
    info.addfrom(n,'Graphic','')
    info.add_deldate(n)
    return info
    
def IT251(n):
    info = Info()
    info.add_deldate(n)
    return info
    
    

class RowHandlers:
  
    @staticmethod
    def IT000(n,name):
        group = _("National Number")
        #~ group = name
        n = n.NationalNumber
        info = Info(
          E.b(n.NationalNumber),
          ' ('+unicode(cbss2gender(n.Sex))+')')
        yield datarow(group,n,n.Date,info)


    @staticmethod
    def FileOwner(fo,name):
        group = _("Residences")
        for n in fo.Residences:
            info = Residence(n)
            yield datarow(group,n,n.Date,info)
            group = ''
            
    @staticmethod
    def AscertainedLegalMainAddresses(fo,name):
        group = _("Ascertained Legal Main Addresses") # Détermination de résidence
        #~ raise Exception(str(fo))
        #~ raise Exception(repr([n for n in fo]))
        for n in fo.AscertainedLegalMainAddress:
            info = IT003(n)
            yield datarow(group,n,n.Date,info)
            group = ''
            
    @staticmethod
    def AddressChangeIntention(fo,name):
        group = _("Address Change Intention") # Intention de changer l'adresse
        for n in fo.Address:
            info = IT005(n)
            yield datarow(group,n,n.Date,info)
            group = ''
            
    @staticmethod
    def AddressReferences(fo,name):
        group = _("Address References") # Adresse de référence
        for n in fo.AddressReference:
            info = IT024(n)
            yield datarow(group,n,n.Date,info)
            group = ''
            

      
    @staticmethod
    def Names(node,name):
        group = _("Names")
        #~ group = name
        for n in node.Name:
            info = Info().addfrom(n,'Name','',NameType)
            yield datarow(group,n,n.Date,info)
            group = ''
        

    @staticmethod
    def LegalMainAddresses(node,name):
        group = _("Legal Main Addresses")
        for n in node.LegalMainAddress:
            info = Info()
            info.chunks.append(E.b(n.Address.ZipCode))
            info.chunks.append(', ')
            info.chunks.append(n.Address.Street.Label)
            info.chunks.append(' ')
            info.chunks.append(n.Address.HouseNumber)
            yield datarow(group,n,n.Date,info)
            group = ''
            

    @staticmethod
    def ResidenceAbroad(node,name):
        def ResidenceAbroadAddressType(n):
            info = Info('Address')
            info.addfrom(n,'PosteDiplomatique',None,DiplomaticPostType)
            info.addfrom(n,'Territory',' ',TerritoryType)
            info.addfrom(n,'Address',' ',AddressType)
            return info
        group = _("Residence Abroad")
        for n in node.ResidenceAbroad:
            info = Info()
            info.addfrom(n,'Address','',ResidenceAbroadAddressType)
            
            #~ info += code_label(n.Address.PosteDiplomatique)
            #~ info.append(', ')
            #~ info += code_label(n.Address.Territory)
            #~ info.append(', ')
            info.add_deldate(n)
            yield datarow(group,n,n.Date,info)
            group = ''
        
    @staticmethod
    def Nationalities(node,name):
        group = _("Nationalities")
        for n in node.Nationality:
            info = code_label(n.Nationality)
            yield datarow(group,n,n.Date,info)
            group = ''
            
    @staticmethod
    def Occupations(node,name):
        group = _("Occupations")
        for n in node.Occupation:
            info = code_label(n.Occupation)
            info.addfrom(n,'SocialCategory',' (SC ',code_label,')')
            #~ info.append(' (SC ')
            #~ info += code_label(n.SocialCategory)
            #~ info.append(')')
            yield datarow(group,n,n.Date,info)
            group = ''
        
    @staticmethod
    def IT100(n,name):
        group = _("Birth Place")
        #~ n = res.BirthPlace
        #~ info = code_label(n.Place1)
        #~ info.append(' (' + n.ActNumber + ')')
        info = Info()
        info.addfrom(n,'Place1',_('in '),PlaceType)
        info.addfrom(n,'Place2',_('in '),GraphicPlaceType)
        info.addfrom(n,'ActNumber',_("Act no. "))
        info.addfrom(n,'SuppletoryRegister')
        yield datarow(group,n,n.Date,info)
        
    @staticmethod
    def IT101(n,name):
        group = _("Declared Birth Date") # Date de naissance déclarée
        info = Info()
        info.addfrom(n,'DeclaredBirthDate','',DateType)
        info.addfrom(n,'Certificate','',CertificateType)
        info.add_deldate(n)
        yield datarow(group,n,n.Date,info)
        
        
    @staticmethod
    def Filiations(node,name):
        group = _("Filiations")
        for n in node.Filiation:
            info = IT110(n)
            yield datarow(group,n,n.Date,info)
            group = ''
          
        
    @staticmethod
    def CivilStates(node,name):
        group = _("Civil States") # IT120
        for n in node.CivilState:
            info = code_label(n.CivilState)
            if hasattr(n,'Spouse'):
                #~ info.append(' with ')
                #~ info += name2info(n.Spouse.Name)
                info.addfrom(n.Spouse,'Name',_('with '),NameType)
                info.chunks.append(' (')
                info.chunks.append(n.Spouse.NationalNumber.NationalNumber)
                info.chunks.append(')')
            info.addfrom(n,'Lieu',_('in '),LieuType)
            #~ info += LieuType(n.Lieu)
            info.addfrom(n,'ActNumber',_("Act no. "))
            #~ info.addfrom(n,'ActNumber')
            info.addfrom(n,'SuppletoryRegister')
            info.add_deldate(n)
            yield datarow(group,n,n.Date,info)
            group = ''
        
    @staticmethod
    def FamilyMembers(node,name):
        group = _("Family Members")
        for n in node.FamilyMember:
            info = IT140(n)
            yield datarow(group,n,n.Date,info)
            group = ''
            
    @staticmethod
    def HeadOfFamily(node,name):
        group = _("Head Of Family")
        for n in node.HeadOfFamily:
            info = Info()
            info.addfrom(n,'FamilyRole','',code_label)
            info.addfrom(n,'Name',_('in family headed by '),NameType)
            #~ info += name2info(n.Name)
            info.chunks.append(' (')
            info.chunks.append(n.NationalNumber.NationalNumber)
            info.chunks.append(')')
            info.add_deldate(n)
            yield datarow(group,n,n.Date,info)
            group = ''
        
    @staticmethod
    def DrivingLicensesOldModel(node,name):
        group = _("Driving Licenses Old Model")
        for n in node.DrivingLicense:
            info = IT194(n)
            yield datarow(group,n,n.Date,info)
            group = ''
        
    @staticmethod
    def DrivingLicenses(node,name):
        group = _("Driving Licenses")
        for n in node.DrivingLicense:
            info = IT191(n)
            yield datarow(group,n,n.Date,info)
            group = ''
        
    @staticmethod
    def WorkPermits(node,name):
        group = _("Work Permits") # Permis de travail
        for n in node.WorkPermit:
            info = IT198(n)
            yield datarow(group,n,n.Date,info)
            group = ''
        
    @staticmethod
    def PublicSecurityNumbers(node,name):
        group = _("Public Security Numbers") # No de securite publique
        for n in node.PublicSecurityNumber:
            info = IT200(n)
            yield datarow(group,n,n.Date,info)
            group = ''
        
    @staticmethod
    def SpecialInfos(node,name):
        group = _("Special Infos") 
        for n in node.SpecialInfo:
            info = IT202(n)
            yield datarow(group,n,n.Date,info)
            group = ''
        
    @staticmethod
    def RefugeeTypes(node,name):
        group = _("Refugee Types") # Type de Personne dans le registre d'attente
        for n in node.RefugeeType:
            info = IT205(n)
            yield datarow(group,n,n.Date,info)
            group = ''
        
    @staticmethod
    def StatusOfRefugee(node,name):
        group = _("Status of refugee") # Statut de refugie
        for n in node.StatusOfRefugee:
            info = IT206(n)
            yield datarow(group,n,n.Date,info)
            group = ''
        
    @staticmethod
    def IdentityCards(node,name):
        group = _("Identity Cards")
        for n in node.IdentityCard:
            info = code_label(n.TypeOfCard)
            info.chunks.append(' ')
            info.chunks.append(_('no. '))
            info.chunks.append(E.b(n.CardNumber))
            info.addfrom(n,'ExpiryDate',_('expires '),DateType)
            #~ info.chunks.append(E.b(dtos(rn2date(n.ExpiryDate))))
            info.addfrom(n,'Delivery',_('delivered in '),DeliveryType)
            #~ info.chunks.append(', delivered in ')
            #~ info += code_label(n.Delivery.Place)
            yield datarow(group,n,n.Date,info)
            group = ''
        
    @staticmethod
    def LegalCohabitations(node,name):
        def CessationType(n):
            info = Info()
            info.addfrom(n,'Reason',_("Reason"),ReasonType)
            info.addfrom(n,'Place',_('in '),PlaceType)
            info.addfrom(n,'Notification',_('in '),NotificationType)
            return info
      
        def DeclarationType(n):
            info = Info()
            info.addfrom(n,'RegistrationDate','',DateType)
            info.addfrom(n,'Partner',_('with '),PartnerType)
            info.addfrom(n,'Place',_('in '),PlaceType)
            info.addfrom(n,'Notary',_('in '),NotaryType)
            return info
    
        group = _("Legal cohabitations")
        for n in node.LegalCohabitation:
            info = Info()
            info.addfrom(n,'Declaration',_("Declaration"),DeclarationType)
            info.addfrom(n,'Cessation',_("Cessation"),CessationType)
            info.add_deldate(n)
            yield datarow(group,n,n.Date,info)
            group = ''
            
      
    @staticmethod
    def Passports(node,name):
        group = _("Passports")
        for n in node.Passport:
            info = IT199(n)
            yield datarow(group,n,n.Date,info)
            group = ''
        
    @staticmethod
    def OrganizationsInCharge(node,name):
        group = _("Organizations in charge")
        for n in node.OrganizationInCharge:
            info = IT207(n)
            yield datarow(group,n,n.Date,info)
            group = ''
        
    @staticmethod
    def RegistrationRegisters(node,name):
        group = _("Registration registers")
        for n in node.RegistrationRegister:
            info = IT210(n)
            yield datarow(group,n,n.Date,info)
            group = ''
        
    @staticmethod
    def ChoosenResidences(node,name):
        group = _("Choosen residences")
        for n in node.ChoosenResidence:
            info = IT212(n)
            yield datarow(group,n,n.Date,info)
            group = ''
        
    @staticmethod
    def OrganDonations(node,name):
        group = _("Organ Donations")
        for n in node.OrganDonation:
            info = Info().addfrom(n,'Declaration','',DeclarationType)
            info.addfrom(n,'Place',_('in '),PlaceType)
            info.add_deldate(n)
            yield datarow(group,n,n.Date,info)
            group = ''
        
    @staticmethod
    def ResidenceUpdateDates(node,name):
        group = _("Residence Update Dates") # Date mise à jour de la résidence principale
        for n in node.ResidenceUpdateDate:
            info = IT251(n)
            yield datarow(group,n,n.Date,info)
            group = ''
        
    @staticmethod
    def DocumentTypes(node,name):
        group = _("Document Types") # Type de document pour déterminer identité
        for n in node.DocumentType:
            info = IT211(n)
            yield datarow(group,n,n.Date,info)
            group = ''
        
    @staticmethod
    def CountriesOfOrigin(node,name):
        group = _("Countries Of Origin") # Pays d'origine
        for n in node.CountryOfOrigin:
            info = IT006(n)
            yield datarow(group,n,n.Date,info)
            group = ''
        
    @staticmethod
    def AddressDeclarationAbroad(node,name):
        group = _("Address Declaration Abroad") 
        for n in node.Address:
            info = IT018(n)
            yield datarow(group,n,n.Date,info)
            group = ''
        
    @staticmethod
    def IT253(node,name):
        group = _("Creation Date")
        n = node # res.CreationDate
        info = Info()
        yield datarow(group,n,n.Date,info)
        
    @staticmethod
    def IT254(node,name):
        group = _("Last Update")
        n = node # res.LastUpdateDate
        info = Info()
        yield datarow(group,n,n.Date,info)
        
    
class RetrieveTIGroupsResult(dd.VirtualTable):
    """
    Displays the response of an :class:`RetrieveTIGroupsRequest`
    as a table.
    """
    master = RetrieveTIGroupsRequest
    master_key = None
    label = _("Results")
    column_names = 'group:18 type:5 since:14 info:50'
    
    @dd.displayfield(_("Group"))
    def group(self,obj,ar):
        return obj.group
            
    @dd.displayfield(_("TI"))
    def type(self,obj,ar):
        return obj.type
            
    @dd.virtualfield(models.DateField(_("Since")))
    def since(self,obj,ar):
        return obj.since
            
    @dd.displayfield(_("Info"))
    def info(self,obj,ar):
        return obj.info
            
    @classmethod
    def get_data_rows(self,ar):
        try:
            rti = ar.master_instance
            if rti is None: 
                #~ print "20120606 ipr is None"
                return
            #~ if not ipr.status in (RequestStates.ok,RequestStates.fictive):
            #~ if not rti.status in (RequestStates.ok,RequestStates.warnings):
                #~ return
            reply = rti.get_service_reply()
            if reply is None:
                return
                
                
            res = reply.rrn_it_implicit
            
            for name, node in res:
                #~ print name, node.__class__
                m = getattr(RowHandlers,node.__class__.__name__,None)
                if m is None:
                    yield AttrDict(
                      info="No handler for %s (%s) in %s" % (
                          name, node.__class__.__name__,rti),
                      group='Error',
                      type='',
                      since=datetime.date.today(),
                      )
                else:
                    for row in m(node,name): yield row
        except Exception, e:
            yield AttrDict(
              info=traceback.format_exc(e),
              group='Traceback',
              type='',
              since=datetime.date.today(),
              )
            
        
        
        
            
__all__ = [
  'RetrieveTIGroupsRequest', 
  'RetrieveTIGroupsRequests',
  'RetrieveTIGroupsRequestsByPerson',
  'MyRetrieveTIGroupsRequests',
  'RetrieveTIGroupsResult',
]


# -*- coding: UTF-8 -*-
## Copyright 2012 Luc Saffre
## This file is part of the Lino project.
## Lino is free software; you can redistribute it and/or modify 
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 3 of the License, or
## (at your option) any later version.
## Lino is distributed in the hope that it will be useful, 
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
## GNU General Public License for more details.
## You should have received a copy of the GNU General Public License
## along with Lino; if not, see <http://www.gnu.org/licenses/>.

import decimal
from dateutil.relativedelta import relativedelta
ONE_DAY = relativedelta(days=1)

from django.db import models
from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.utils.translation import ugettext as _


from lino.utils import i2d
from lino.utils.instantiator import Instantiator
from lino.core.modeltools import resolve_model
from lino.utils.babel import babel_values

from lino.modlib.debts.models import AccountType

def n2dec(v):
    return decimal.Decimal("%.2d" % v)
   
def objects():
    group = Instantiator('debts.AccountGroup').build
    g = group(account_type=AccountType.income,**babel_values('name',
          de=u"Monatliche Einkünfte",
          fr=u"Revenus mensuels",
          en=u"Monthly incomes"
          ))
    yield g
    account = Instantiator('debts.Account',group=g).build
    yield account(required_for_person=True,**babel_values('name',
          de=u"Gehälter",
          fr=u"Salaires",
          en=u"Salaries"
          ))
    yield account(required_for_person=True,**babel_values('name',
          de=u"Renten",
          fr=u"Pension",
          en=u"Pension"
          ))
    yield account(required_for_person=True,**babel_values('name',
          de=u"Integrationszulage",
          fr=u"Allocation d'intégration",
          en=u"Integration aid"
          ))
    yield account(required_for_person=True,**babel_values('name',
          de=u"Ersatzeinkünfte",
          fr=u"Ersatzeinkünfte",
          en=u"Ersatzeinkünfte"
          ))
    yield account(required_for_person=True,**babel_values('name',
          de=u"Alimente",
          fr=u"Aliments",
          en=u"Aliments"
          ))
    yield account(required_for_person=True,**babel_values('name',
          de=u"Essen-Schecks",
          fr=u"Chèques-repas",
          en=u"Chèques-repas"
          ))
    yield account(required_for_person=True,**babel_values('name',
          de=u"Andere",
          fr=u"Andere",
          en=u"Andere"
          ))

    g = group(account_type=AccountType.income,**babel_values('name',
          de=u"Jährliche Einkünfte",
          fr=u"Revenus annuels",
          en=u"Yearly incomes"
          ))
    yield g
    account = Instantiator('debts.Account',group=g,periods=12).build
    yield account(required_for_person=True,**babel_values('name',
          de=u"Urlaubsgeld",
          fr=u"Congé payé",
          en=u"Paid holiday"
          ))
    yield account(required_for_person=True,**babel_values('name',
          de=u"Jahresendzulage",
          fr=u"Prime de fin d'année",
          en=u"Year-end prime"
          ))
    yield account(required_for_person=True,**babel_values('name',
          de=u"Gewerkschaftsprämie",
          fr=u"Gewerkschaftsprämie",
          en=u"Gewerkschaftsprämie"
          ))

    g = group(account_type=AccountType.expense,**babel_values('name',
          de=u"Monatliche Ausgaben",
          fr=u"Dépenses mensuelles",
          en=u"Monthly expenses"
          ))
    yield g
    account = Instantiator('debts.Account',group=g).build
    yield account(required_for_household=True,**babel_values('name',
          de=u"Miete",
          fr=u"Loyer",
          en=u"Rent"
          ))
    yield account(required_for_household=True,**babel_values('name',
          de=u"Wasser",
          fr=u"Eau",
          en=u"Water"
          ))
    yield account(required_for_household=True,**babel_values('name',
          de=u"Strom",
          fr=u"Electricité",
          en=u"Electricity"
          ))
    yield account(required_for_household=True,**babel_values('name',
          de=u"Festnetz-Telefon und Internet",
          fr=u"Téléphone fixe et Internet",
          en=u"Telephone & Internet"
          ))
    yield account(required_for_household=True,**babel_values('name',
          de=u"Handy",
          fr=u"GSM",
          en=u"Cell phone"
          ))
    yield account(required_for_household=True,**babel_values('name',
          de=u"TEC Busabonnement",
          fr=u"Abonnement bus",
          en=u"Public transport"
          ))
    yield account(required_for_household=True,**babel_values('name',
          de=u"Fahrtkosten",
          fr=u"Frais de transport",
          en=u"Transport costs"
          ))
    yield account(required_for_household=True,**babel_values('name',
          de=u"Benzin",
          fr=u"Essence",
          en=u"Fuel"
          ))
    yield account(required_for_household=True,**babel_values('name',
          de=u"Unterhalt Auto",
          fr=u"Maintenance voiture",
          en=u"Car maintenance"
          ))
    yield account(required_for_household=True,**babel_values('name',
          de=u"Schulkosten",
          fr=u"École",
          en=u"School"
          ))
    yield account(required_for_household=True,**babel_values('name',
          de=u"Tagesmutter & Kleinkindbetreuung",
          fr=u"Garde enfant",
          en=u"Babysitting"
          ))
    yield account(required_for_household=True,**babel_values('name',
          de=u"Gesundheit",
          fr=u"Santé",
          en=u"Health"
          ))
    yield account(required_for_household=True,**babel_values('name',
          de=u"Kleidung",
          fr=u"Vêtements",
          en=u"Clothes"
          ))
    yield account(required_for_household=True,**babel_values('name',
          de=u"Ernährung",
          fr=u"Alimentation",
          en=u"Food"
          ))
    yield account(required_for_household=True,**babel_values('name',
          de=u"Hygiene",
          fr=u"Hygiène",
          en=u"Hygiene"
          ))
    yield account(required_for_household=True,**babel_values('name',
          de=u"Krankenkassenbeiträge",
          fr=u"Mutuelle",
          en=u"Health insurance"
          ))
    yield account(required_for_household=True,**babel_values('name',
          de=u"Gewerkschaftsbeiträge",
          fr=u"Cotisations syndicat",
          en=u"Labour fees"
          ))
    yield account(required_for_household=True,**babel_values('name',
          de=u"Unterhaltszahlungen",
          fr=u"Unterhaltszahlungen",
          en=u"Unterhaltszahlungen"
          ))
    yield account(required_for_household=True,**babel_values('name',
          de=u"Tabak",
          fr=u"Tabac",
          en=u"Tobacco"
          ))
    yield account(required_for_household=True,**babel_values('name',
          de=u"Freizeit & Unterhaltung",
          fr=u"Loisirs",
          en=u"Spare time"
          ))
    yield account(required_for_household=True,**babel_values('name',
          de=u"Haustiere",
          fr=u"Animaux domestiques",
          en=u"Pets"
          ))
    yield account(required_for_household=True,**babel_values('name',
          de=u"Pensionssparen",
          fr=u"Épargne pension",
          en=u"Retirement savings"
          ))
    yield account(required_for_household=True,**babel_values('name',
          de=u"Sonstige",
          fr=u"Autres",
          en=u"Other"
          ))


    g = group(account_type=AccountType.expense,**babel_values('name',
          de=u"Steuern",
          fr=u"Taxes",
          en=u"Taxes"
          ))
    yield g
    account = Instantiator('debts.Account',group=g,periods=12).build
    yield account(required_for_household=True,**babel_values('name',
          de=u"Gemeindesteuer",
          fr=u"Taxe communale",
          en=u"Municipal tax"
          ))
    yield account(required_for_household=True,**babel_values('name',
          de=u"Kanalisationssteuer",
          fr=u"Kanalisationssteuer",
          en=u"Kanalisationssteuer"
          ))
    yield account(required_for_household=True,**babel_values('name',
          de=u"Müllsteuer",
          fr=u"Taxe déchets",
          en=u"Waste tax"
          ))
    yield account(required_for_household=True,**babel_values('name',
          de=u"Autosteuer",
          fr=u"Taxe circulation",
          en=u"Autosteuer"
          ))
    yield account(required_for_household=True,**babel_values('name',
          de=u"Immobiliensteuer",
          fr=u"Taxe immobilière",
          en=u"Immobiliensteuer"
          ))
    yield account(required_for_household=True,**babel_values('name',
          de=u"Andere",
          fr=u"Autres",
          en=u"Other"
          ))

    g = group(account_type=AccountType.expense,**babel_values('name',
          de=u"Versicherungen",
          fr=u"Assurances",
          en=u"Insurances"
          ))
    yield g
    account = Instantiator('debts.Account',group=g,periods=12).build
    yield account(required_for_household=True,**babel_values('name',
          de=u"Feuer",
          fr=u"Incendie",
          en=u"Fire"
          ))
    yield account(required_for_household=True,**babel_values('name',
          de=u"Familienhaftpflicht",
          fr=u"Responsabilité famille",
          en=u"Familienhaftpflicht"
          ))
    yield account(required_for_household=True,**babel_values('name',
          de=u"Auto",
          fr=u"Voiture",
          en=u"Car insurance"
          ))
    yield account(required_for_household=True,**babel_values('name',
          de=u"Lebensversicherung",
          fr=u"Assurance vie",
          en=u"Life insurance"
          ))
    yield account(required_for_household=True,**babel_values('name',
          de=u"Andere Versicherungen",
          fr=u"Autres assurances",
          en=u"Other insurances"
          ))
          
          
    g = group(account_type=AccountType.asset,**babel_values('name',
          de=u"Aktiva, Vermögen, Kapital",
          fr=u"Actifs",
          en=u"Assets"
          ))
    yield g
    account = Instantiator('debts.Account',group=g).build
    yield account(**babel_values('name',
          de=u"Haus",
          fr=u"Maison",
          en=u"House"
          ))
    yield account(**babel_values('name',
          de=u"Auto",
          fr=u"Voiture",
          en=u"Car"
          ))
    
    
    g = group(account_type=AccountType.liability,**babel_values('name',
          de=u"Guthaben, Schulden, Verbindlichkeit",
          fr=u"Créances et dettes",
          en=u"Liabilities"
          ))
    yield g
    account = Instantiator('debts.Account',group=g).build
    yield account(**babel_values('name',
          de=u"Kredite",
          fr=u"Crédits",
          en=u"Loans"
          ))
    yield account(**babel_values('name',
          de=u"Schulden",
          fr=u"Emprunts",
          en=u"Debts"
          ))
    yield account(**babel_values('name',
          de=u"Gerichtsvollzieher",
          fr=u"Huissier de justice",
          en=u"Bailiff"
          ))
    yield account(**babel_values('name',
          de=u"Zahlungsrückstände",
          fr=u"Factures à payer",
          en=u"Invoices to pay"
          ))




# -*- coding: UTF-8 -*-
## Copyright 2008-2012 Luc Saffre
## This file is part of the Lino project.
## Lino is free software; you can redistribute it and/or modify 
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 3 of the License, or
## (at your option) any later version.
## Lino is distributed in the hope that it will be useful, 
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
## GNU General Public License for more details.
## You should have received a copy of the GNU General Public License
## along with Lino; if not, see <http://www.gnu.org/licenses/>.

import logging
logger = logging.getLogger(__name__)

import os
import cgi
import datetime

from django.db import models
from django.db.models import Q
from django.db.utils import DatabaseError
from django.conf import settings
from django.core.exceptions import ValidationError
from django.utils.safestring import mark_safe
from django.utils.translation import ugettext_lazy as _
from django.utils.translation import string_concat
from django.utils.encoding import force_unicode 
from django.utils.functional import lazy
    

#~ import lino
#~ logger.debug(__file__+' : started')
#~ from django.utils import translation


#~ from lino import reports
from lino import dd
#~ from lino import layouts
#~ from lino.utils import printable
from lino import mixins
#~ from lino import actions
#~ from lino import fields
from lino.modlib.contacts import models as contacts
from lino.modlib.notes import models as notes
#~ from lino.modlib.links import models as links
from lino.modlib.uploads import models as uploads
from lino.modlib.cal import models as cal
#~ from lino.modlib.users import models as users
from lino.utils.choicelists import HowWell, Gender
from lino.utils.choicelists import ChoiceList
#~ from lino.modlib.properties.utils import KnowledgeField #, StrengthField
#~ from lino.modlib.uploads.models import UploadsByPerson
from lino.core.modeltools import get_field
from lino.core.modeltools import resolve_field
from lino.core.modeltools import range_filter
from lino.utils.babel import DEFAULT_LANGUAGE, babelattr, babeldict_getitem
from lino.utils.babel import language_choices
#~ from lino.utils.babel import add_babel_field, DEFAULT_LANGUAGE, babelattr, babeldict_getitem
from lino.utils import babel 
from lino.utils.choosers import chooser
from lino.utils import mti
from lino.mixins.printable import DirectPrintAction, Printable
#~ from lino.mixins.reminder import ReminderEntry
from lino.core.modeltools import obj2str

from lino.modlib.countries.models import CountryCity
from lino.modlib.properties import models as properties
from lino.modlib.cal.models import update_reminder
#~ from lino.modlib.contacts.models import Contact
from lino.core.modeltools import resolve_model, UnresolvedModel

#~ # not used here, but these modules are required in INSTALLED_APPS, 
#~ # and other code may import them using 
#~ # ``from lino.apps.pcsw.models import Property``

#~ from lino.modlib.properties.models import Property
#~ # from lino.modlib.notes.models import NoteType
#~ from lino.modlib.countries.models import Country, City

#~ if settings.LINO.user_model:
    #~ User = resolve_model(settings.LINO.user_model,strict=True)



class CefLevel(ChoiceList):
    """
    Levels of the Common European Framework (CEF).
    
    | http://www.coe.int/t/dg4/linguistic/CADRE_EN.asp
    | http://www.coe.int/t/dg4/linguistic/Source/ManualRevision-proofread-FINAL_en.pdf
    | http://www.telc.net/en/what-telc-offers/cef-levels/a2/
    
    """
    label = _("CEF level")
    show_values = True
    
    #~ @classmethod
    #~ def display_text(cls,bc):
        #~ def fn(bc):
            #~ return u"%s (%s)" % (bc.value,unicode(bc))
        #~ return lazy(fn,unicode)(bc)
    
add = CefLevel.add_item
add('A1', _("basic language skills"))
add('A2', _("basic language skills"))
add('A2+', _("basic language skills"))
add('B1', _("independent use of language"))
add('B2', _("independent use of language"))
add('B2+', _("independent use of language"))
add('C1', _("proficient use of language"))
add('C2', _("proficient use of language"))
add('C2+', _("proficient use of language"))





#
# LanguageKnowledge
#

class LanguageKnowledge(dd.Model):
    """Specifies how well a certain Person knows a certain Language.
    Deserves more documentation."""
    class Meta:
        verbose_name = _("language knowledge")
        verbose_name_plural = _("language knowledges")
        
    allow_cascaded_delete = True
    
    #~ person = models.ForeignKey("contacts.Person")
    person = models.ForeignKey(settings.LINO.person_model)
    language = models.ForeignKey("countries.Language",verbose_name=_("Language"))
    #~ language = models.ForeignKey("countries.Language")
    #~ language = fields.LanguageField()
    spoken = HowWell.field(verbose_name=_("spoken"))
    written = HowWell.field(verbose_name=_("written"))
    native = models.BooleanField(verbose_name=_("native language"))
    cef_level = CefLevel.field(blank=True) # ,null=True)
    
    def __unicode__(self):
        if self.language_id is None:
            return ''
        if self.cef_level:
            return u"%s (%s)" % (self.language,self.cef_level)
        if self.spoken > '1' and self.written > '1':
            return _(u"%s (s/w)") % self.language
        elif self.spoken > '1':
            return _(u"%s (s)") % self.language
        elif self.written > '1':
            return _(u"%s (w)") % self.language
        else:
            return unicode(self.language)
      
    
class LanguageKnowledgesByPerson(dd.Table):
    model = LanguageKnowledge
    master_key = 'person'
    #~ label = _("Language knowledge")
    #~ button_label = _("Languages")
    column_names = "language native spoken written cef_level"

# 
# PROPERTIES
# 


class PersonProperty(properties.PropertyOccurence):
    """A given property defined for a given person. 
    See :mod:`lino.modlib.properties`."""
    
    allow_cascaded_delete = True
    
    class Meta:
        app_label = 'properties'
        verbose_name = _("Property")
        verbose_name_plural = _("Properties")
        
    #~ person = models.ForeignKey("contacts.Person")
    person = models.ForeignKey(settings.LINO.person_model)
    remark = models.CharField(max_length=200,
        blank=True,# null=True,
        verbose_name=_("Remark"))


class PersonProperties(dd.Table):
    model = PersonProperty
    hidden_columns = frozenset(['group'])
    
class PropsByPerson(PersonProperties):
    master_key = 'person'
    column_names = "property value remark *"
    
    
class PersonPropsByProp(PersonProperties):
    model = PersonProperty
    #~ app_label = 'properties'
    master_key = 'property'
    column_names = "person value remark *"
    
#~ class PersonPropsByType(dd.Table):
    #~ model = PersonProperty
    #~ master_key = 'type'
    #~ column_names = "person property value remark *"
    #~ hidden_columns = frozenset(['group'])
    
class ConfiguredPropsByPerson(PropsByPerson):
    """
    Base class for 
    :class`SkillsByPerson`, 
    :class`SoftSkillsByPerson` and
    :class`ObstaclesByPerson`.
    """
    propgroup_config_name = None
    typo_check = False # to avoid warning "ConfiguredPropsByPerson 
                       # defines new attribute(s) propgroup_config_name"
    #~ def setup_actions(self):
    @classmethod
    def class_init(self):
        if self.propgroup_config_name:
            pg = getattr(settings.LINO.site_config,self.propgroup_config_name)
            self.known_values = dict(group=pg)
            if pg is None:
                self.label = _("(Site setting %s is empty)" % self.propgroup_config_name)
            else:
                #~ def f():
                    #~ return babelattr(pg,'name')
                #~ self.label = lazy(f,unicode)()
                self.label = lazy(babelattr,unicode)(pg,'name')
                #~ self.label = babelattr(pg,'name')
        super(PropsByPerson,self).class_init()
        
class SkillsByPerson(ConfiguredPropsByPerson):
    propgroup_config_name = 'propgroup_skills'
        
class SoftSkillsByPerson(ConfiguredPropsByPerson):
    propgroup_config_name = 'propgroup_softskills'
        
class ObstaclesByPerson(ConfiguredPropsByPerson):
    propgroup_config_name = 'propgroup_obstacles'
    
    




"""
Here we add some specific fields to the 
standard model SiteConfig.
http://osdir.com/ml/django-users/2009-11/msg00696.html
"""

from lino.models import SiteConfig
    
dd.inject_field(SiteConfig,
    'propgroup_skills',
    models.ForeignKey('properties.PropGroup',
        blank=True,null=True,
        verbose_name=_("Skills Property Group"),
        related_name='skills_sites'),
    """The property group to be used as master 
    for the SkillsByPerson report.""")
dd.inject_field(SiteConfig,
    'propgroup_softskills',
    models.ForeignKey('properties.PropGroup',
        blank=True,null=True,
        verbose_name=_("Soft Skills Property Group"),
        related_name='softskills_sites',
        ),
    """The property group to be used as master 
    for the SoftSkillsByPerson report."""
    )
dd.inject_field(SiteConfig,
    'propgroup_obstacles',
    models.ForeignKey('properties.PropGroup',
        blank=True,null=True,
        verbose_name=_("Obstacles Property Group"),
        related_name='obstacles_sites',
        ),
    """The property group to be used as master 
    for the ObstaclesByPerson report."""
    )






# -*- coding: UTF-8 -*-
## Copyright 2011-2012 Luc Saffre
## This file is part of the Lino project.
## Lino is free software; you can redistribute it and/or modify 
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 3 of the License, or
## (at your option) any later version.
## Lino is distributed in the hope that it will be useful, 
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
## GNU General Public License for more details.
## You should have received a copy of the GNU General Public License
## along with Lino; if not, see <http://www.gnu.org/licenses/>.

u"""
Adds models and tables used to make the following :term:`CBSS` requests
(descriptions taken from `SYNTHESE_SERVICESXML_CPAS_FR.DOC`):

- :class:`IdentifyPersonRequest` :
  Identifier la personne par son NISS ou ses données phonétiques et vérifier son identité par le numéro de carte SIS, de carte d'identité ou par ses données phonétiques ()

- :class:`ManageAccessRequest`: 
  Enregistrer, désenregistrer ou consulter (:class:`ManageAction`) un dossier 
  dans le registre du réseau de la sécurité sociale (registre BCSS) 
  et dans le répertoire sectoriel des CPAS géré par la SmalS-MvM (:class:`QueryRegister`).
  
  - `ManageAction.REGISTER` : 
    Ce service est sollicité au moment du démarrage de l’enquête sociale.  
    Le CPAS déclare au réseau de la sécurité sociale qu’il possède un dossier pour lequel il a 
    l’autorisation (dispositions légales et réglementaires) d’obtenir des informations des autres 
    institutions en vue de compléter son enquête dans le cadre de l’octroi du revenu d’intégration.  
    Cette déclaration concerne le répertoire sectoriel des CPAS à la SmalS-MvM et peut 
    concerner plusieurs catégories de personnes : 
    le demandeur, les cohabitants et les tiers concernés 
    et ce, pour des finalités différentes. 
  - `ManageAction.UNREGISTER` : 
    L’opération contraire est aussi mise à disposition. 
  - `ManageAction.LIST` : 
    Il est en plus possible d’obtenir une liste des enregistrements 
    dans le répertoire sectoriel des CPAS à la SmalS-MvM 
    ainsi qu’au sein du réseau BCSS.
  
- :class:`RetrieveTIGroupsRequest`: 
  Obtenir des informations à propos d’une personne dans le cadre de l’enquête sociale.

"""

import os
import shutil
import traceback
import datetime
import logging
logger = logging.getLogger(__name__)


from django.conf import settings
from django.db import models
from django.utils.translation import ugettext_lazy as _

from appy.shared.xml_parser import XmlUnmarshaller


from lino import mixins
from lino import dd
from lino.utils import Warning
from lino.utils import join_words
from lino.utils import AttrDict, IncompleteDate
from lino.core.modeltools import obj2str

from lino.utils import babel
from lino.utils.choosers import chooser

from lino.utils.xmlgen import html as xghtml

#~ from lino.utils import dblogger
#~ from lino.core.modeltools import resolve_model
#~ from lino.utils.xmlgen import etree
#~ from lino.utils.xmlgen import cbss

from lino.utils.babel import dtos

from lino.utils.choicelists import ChoiceList
from lino.utils.choicelists import Gender
from lino.modlib.users.models import UserLevels
#~ from lino.modlib.contacts import models as contacts
from lino.core.modeltools import makedirs_if_missing
#~ from lino.mixins.printable import DirectPrintAction

from lino.apps.pcsw import models as pcsw

try:
  
    import suds
    from suds.client import Client
    from suds.transport.http import HttpAuthenticated
    from suds.transport.http import HttpTransport
    from suds.sax.element import Element as E
    from suds.sax.parser import Parser
    PARSER = Parser()

except ImportError, e:
    pass



countries = dd.resolve_app('countries')

_clients_dict = dict()

def get_client(obj):
    c = _clients_dict.get(obj.__class__,None)
    if c is not None: return c
    c = obj.create_client()
    _clients_dict[obj.__class__] = c
    return c
    


CBSS_ENVS = ('test', 'acpt', 'prod')

def xsdpath(*parts):
    p1 = os.path.abspath(os.path.dirname(__file__))
    return os.path.join(p1,'XSD',*parts)

CBSS_ERROR_MESSAGE = "CBSS error %s:\n"

def gender2cbss(gender):
    if gender == Gender.male:
        return '1'
    elif gender == Gender.female:
        return '2'
    else:
        return '0'
            

def cbss2gender(v):
    if v == '1':
        return Gender.male
    elif v == '2':
        return Gender.female
    return None
      
def cbss2date(s):
    a = s.split('-')
    assert len(a) == 3
    a = [int(i) for i in a]
    return IncompleteDate(*a)


def cbss2civilstate(node):
    value = nodetext(node)
    if not value:
        return value
    v = pcsw.CivilState.get_by_value(value)
    if v is None:
        print "20120601 cbss2civilstate None for ", repr(value)
    return unicode(v)
    
def nodetext(node):
    if node is None:
        return ''
    return node.text
    
def cbss2country(code):
    try:
        return countries.Country.objects.get(inscode=code)
    except Country.DoesNotExist:
        logger.warning("Unknown country code %s",code)
    
    
def cbss2address(obj,**data):
    n = obj.childAtPath('/Basic/DiplomaticPost')
    if n is not None:
        data.update(country=cbss2country(nodetext(n.childAtPath('/CountryCode'))))
        #~ n.childAtPath('/Post')
        data.update(address=nodetext(n.childAtPath('/AddressPlainText')))
        return data
    n = obj.childAtPath('/Basic/Address')
    if n is not None:
        data.update(country=cbss2country(nodetext(n.childAtPath('/CountryCode'))))
        #~ country = countries.Country.objects.get(
            #~ inscode=n.childAtPath('/CountryCode').text)
        addr = ''
        #~ addr += n.childAtPath('/MunicipalityCode').text
        addr += join_words(
            nodetext(n.childAtPath('/Street')),
            nodetext(n.childAtPath('/HouseNumber')),
            nodetext(n.childAtPath('/Box'))
            )
        addr += ', ' + join_words(
            nodetext(n.childAtPath('/PostalCode')),
            nodetext(n.childAtPath('/Municipality'))
            )
        data.update(address=addr)
    return data
    
    
    
class RequestStates(ChoiceList):
    """
    The status of a :class:`CBSSRequest`.
    """
    label = _("State")

add = RequestStates.add_item
#~ add('0',_("New"),'new')
add('10',_("Sent"),'sent')
add('20',_("Failed"),'failed') # sending failed. no ticket yet. can execute again.
add('25',_("Validated"),'validated') # only when cbss_live_tests
add('30',_("OK"),'ok')
add('40',_("Warnings"),'warnings') # OK and useable
add('50',_("Errors"),'errors') # there's a ticket, but no usable result. cannot print.
#~ add('6',_("Invalid reply"),'invalid')
#~ add('9',_("Fictive"),'fictive')
  
#~ class Environment(ChoiceList):
    #~ """
    #~ The environment where a :class:`CBSSRequest` is being executed.
    #~ """
    #~ label = _("Environment")
#~ add = Environment.add_item
#~ add('t',_("Test"),'test')
#~ add('a',_("Acceptance"),'acpt')
#~ add('p',_("Production"),'prod')
  

    
    


class unused_ExecuteRequest(dd.RowAction):
    """
    This defines the "Execute" button on a 
    :class:`CBSSRequest` or
    :class:`SSDNRequest` 
    record.
    """
    readonly = False
    url_action_name = 'exec'
    label = _('Execute')
    #~ callable_from = None
    callable_from = (dd.GridEdit,dd.ShowDetailAction)
    required = dict(states=['','validated','failed'])
    
    #~ def get_row_permission(self,user,obj):
        #~ if obj.ticket:
            #~ return False
        #~ return super(ExecuteRequest,self).get_row_permission(user,obj)
      
    def run(self,obj,ar,**kw):
        obj.execute_request(ar)
        if obj.status == RequestStates.failed:
            kw.update(message=obj.debug_messages)
            kw.update(alert=True)
        elif obj.status == RequestStates.warnings:
            kw.update(message=obj.info_messages)
            #~ kw.update(message=_("Got valid response, but it contains warnings."))
            kw.update(alert=True)
        kw.update(refresh=True)
        return ar.ui.success_response(**kw)



class Sector(babel.BabelNamed):
    """
    Default values filled from :mod:`lino.modlib.cbss.fixtures.sectors`.
    """
    class Meta:
        verbose_name = _("Sector")
        verbose_name_plural = _('Sectors')
        unique_together = ['code','subcode']
        
    #~ code = models.CharField(max_length=2,verbose_name=_("Code"),primary_key=True)
    code = models.IntegerField(max_length=2,verbose_name=_("Code"))
    subcode = models.IntegerField(max_length=2,verbose_name=_("Subcode"),default=0)
    abbr = babel.BabelCharField(_("Abbreviation"),max_length=50,blank=True)
    
    
    def __unicode__(self):
        #~ return '(' + str(self.code) + ') ' + babel.BabelNamed.__unicode__(self)
        if self.subcode != 0:
            return str(self.code) + '.' + str(self.subcode) + ' - ' + babel.BabelNamed.__unicode__(self)
        return str(self.code) + ' - ' + babel.BabelNamed.__unicode__(self)
        

class Sectors(dd.Table):
    model = Sector
    #~ read_permission = perms.Required(user_groups = ['cbss'])
    #~ required_user_groups = ['cbss']
    required = dict(user_groups='cbss')
    column_names = 'code subcode abbr name *'
    order_by = ['code','subcode']
    

class Purpose(babel.BabelNamed):
    u"""
    Codes qualité (Hoedanigheidscodes),
    http://www.bcss.fgov.be/binaries/documentation/fr/documentation/general/lijst_hoedanigheidscodes.pdf
    """
    class Meta:
        verbose_name = _("Purpose")
        verbose_name_plural = _('Purposes')
        unique_together = ['sector_code','code']
    sector_code = models.IntegerField(max_length=2,verbose_name=_("Sector"),blank=True,null=True)
    #~ sector_subcode = models.IntegerField(max_length=2,verbose_name=_("Subsector"),blank=True,null=True)
    #~ sector = models.ForeignKey(Sector,blank=True,null=True)
    #~ code = models.CharField(max_length=3,verbose_name=_("Code"))
    code = models.IntegerField(max_length=3,verbose_name=_("Code"))
    
    def __unicode__(self):
        #~ return '(' + str(self.code) + ') ' + babel.BabelNamed.__unicode__(self)
        return str(self.code) + ' - ' + babel.BabelNamed.__unicode__(self)
        

class Purposes(dd.Table):
    model = Purpose
    required = dict(user_groups='cbss')
    #~ required_user_groups = ['cbss']
    column_names = 'sector_code code name *'
    order_by = ['sector_code','code']




NSCOMMON = ('common','http://www.ksz-bcss.fgov.be/XSD/SSDN/Common')
NSSSDN = ('ssdn','http://www.ksz-bcss.fgov.be/XSD/SSDN/Service')
NSIPR = ('ipr',"http://www.ksz-bcss.fgov.be/XSD/SSDN/OCMW_CPAS/IdentifyPerson")
NSMAR = ('mar',"http://www.ksz-bcss.fgov.be/XSD/SSDN/OCMW_CPAS/ManageAccess")
NSWSC = ('wsc',"http://ksz-bcss.fgov.be/connectors/WebServiceConnector")


#~ class CBSSRequest(mixins.ProjectRelated,mixins.AutoUser):
class CBSSRequest(mixins.AutoUser,mixins.Printable,mixins.Duplicable):
    """
    Common Abstract Base Class for :class:`SSDNRequest`
    and :class:`NewStyleRequest`
    """
    
    workflow_state_field = 'status'
    
    wsdl_parts = NotImplementedError
    
    class Meta:
        abstract = True
        
    person = models.ForeignKey(
        settings.LINO.person_model,
        verbose_name=_("Client"))
    
    sent = models.DateTimeField(
        verbose_name=_("Sent"),
        blank=True,null=True,
        editable=False,
        help_text="""\
The date and time when this request has been executed. 
This is empty for requests than haven't been sent.
Read-only.""")
    
    #~ status = RequestStates.field(default=RequestStates.new,editable=False)
    status = RequestStates.field(editable=False)
    environment = models.CharField(max_length=4,editable=False,verbose_name=_("T/A/B"))
    ticket  = models.CharField(max_length=36,editable=False,verbose_name=_("Ticket"))
    #~ environment = Environment.field(blank=True,null=True)
    
    # will probably go away soon
    request_xml = models.TextField(verbose_name=_("Request"),
        editable=False,blank=True,
        help_text="""The raw XML string that has (or would have) been sent.""")
    
    response_xml = models.TextField(
        verbose_name=_("Response"),
        editable=False,blank=True,
        help_text="""\
The raw XML response received. 
""")

    #~ logged_messages = models.TextField(
        #~ verbose_name=_("Logged messages"),
        #~ editable=False,blank=True,
        #~ help_text="""Logged messages about this request.""")

    debug_messages = models.TextField(
        verbose_name=_("Debug messages"),
        editable=False,blank=True)
    info_messages = models.TextField(
        verbose_name=_("Info messages"),
        editable=False,blank=True)

    #~ send_action = ExecuteRequest()
    #~ print_action = mixins.DirectPrintAction(required=dict(states=['ok','warnings']))
    print_action = mixins.DirectPrintAction()
    
    
    #~ def save(self,*args,**kw):
        #~ if not self.environment:
            #~ self.environment = settings.LINO.cbss_environment or ''
        #~ super(CBSSRequest,self).save(*args,**kw)

    #~ def disable_editing(self,ar):
        #~ if self.ticket: return True

    def on_duplicate(self,ar,master):
        """
        When duplicating a CBSS request, we want re-execute it. 
        So please duplicate only the parameters, 
        not the execution data like `ticket`, `sent` and `status`.
        Note that also the `user` will be set to the user who asked to duplicate
        (because this is a subclass of :mod:`lino.mixins.UserAuthored`.
        """
        self.debug_messages = ''
        self.info_messages = ''
        self.ticket = ''
        self.response_xml = ''
        self.request_xml = ''
        self.sent = None
        #~ self.status = RequestStates.new
        self.status = RequestStates.blank_item
        self.environment = ''
        super(CBSSRequest,self).on_duplicate(ar,master)
        
    def get_row_permission(self,user,state,action):
        """
        CBSS requests that have a `ticket` may never be modified.
        """
        #~ logger.info("20120622 CBSSRequest.get_row_permission %s %s", self.ticket, action.readonly)
        if self.ticket and not action.readonly: 
            return False
        return super(CBSSRequest,self).get_row_permission(user,state,action)
        
    def on_cbss_ok(self,reply):
        """
        Called when a successful reply has been received.
        """
        pass
        
    #~ @classmethod
    #~ def setup_report(cls,rpt):
        #~ # call_optional_super(CBSSRequest,cls,'setup_report',rpt)
        #~ rpt.add_action(ExecuteRequest())
        
    #~ def logmsg(self,s,*args):
        #~ if args:
            #~ s = s % args
        #~ self.logged_messages += ("[%s] " % datetime.datetime.now()) + s + '\n'
        
    def logmsg_debug(self,s,*args):
        if args:
            s = s % args
        self.debug_messages += ("[%s] " % datetime.datetime.now()) + s + '\n'
        
    def logmsg_info(self,s,*args):
        if args:
            s = s % args
        self.info_messages += s + '\n'
        
    def logmsg_warning(self,s,*args):
        if args:
            s = s % args
        self.info_messages += s + '\n'
        
    def __unicode__(self):
        return u"%s #%s" % (self._meta.verbose_name,self.pk)
        
    def after_ui_save(self,ar,**kw):
        self.execute_request(ar)
        if self.status == RequestStates.failed:
            kw.update(message=self.debug_messages)
            kw.update(alert=True)
        elif self.status == RequestStates.warnings:
            kw.update(message=self.info_messages)
            #~ kw.update(message=_("Got valid response, but it contains warnings."))
            kw.update(alert=True)
        #~ kw.update(refresh=True)
        #~ return ar.success_response(**kw)
        return kw

        

    def execute_request(self,ar=None,now=None,simulate_response=None,environment=None):
        """
        This is the general method for all SSDN services,
        executed when a user runs :class:`ExecuteRequest`.
        """
        if self.ticket:
            raise Warning("Cannot re-execute %s with non-empty ticket." % self)
        if ar is not None:
            logger.info("%s executes CBSS request %s",ar.get_user(),self)
        if now is None:
            now = datetime.datetime.now()
        if environment is None:
            environment = settings.LINO.cbss_environment or ''
            
        self.environment = environment
        self.sent = now
        #~ self.logged_messages = ''
        self.debug_messages = ''
        self.info_messages = ''
        
        if not settings.LINO.cbss_live_tests:
            if simulate_response is None and environment:
                self.validate_request()
                self.status = RequestStates.validated
                self.save()
                return
                
        self.status = RequestStates.sent
        self.save()
        
        retval = None
        try:
            #~ if not settings.LINO.cbss_live_tests:
                #~ self.validate_request()
            retval = self.execute_request_(now,simulate_response)
        except (IOError,Warning),e:
            if self.ticket:
                self.status = RequestStates.errors
            else:
                self.status = RequestStates.failed
            self.logmsg_debug(unicode(e))
        except Exception,e:
            if self.ticket:
                self.status = RequestStates.errors
            else:
                self.status = RequestStates.failed
            #~ self.response_xml = traceback.format_exc(e)
            self.logmsg_debug(traceback.format_exc(e))

        self.save()
        return retval


    
    #~ @dd.action(_("Validate"))
    #~ def validate(self,ar):
        #~ try:
            #~ self.validate_request()
            #~ self.save()
            #~ return ar.ui.success_response(
                #~ message="%s validation passed." % self)
        #~ except Exception,e:
            #~ self.logmsg(traceback.format_exc(e))
            #~ self.save()
            #~ return ar.ui.error_response(e)
          
        
          
        
        
            
    def validate_request(self):
        pass

    def get_wsdl_uri(self):
        url = os.path.join(settings.MEDIA_ROOT,*self.wsdl_parts) 
        if not url.startswith('/'):
            # on a windows machine we need to prepend an additional "/"
            url = '/' + url
        if os.path.sep != '/':
            url = url.replace(os.path.sep,'/')
        url = 'file://' + url 
        return url

    def check_environment(self,req):
        if not self.environment:
            raise Warning("""\
Not actually sending because environment is empty. Request would be:
""" + unicode(req))

        assert self.environment in CBSS_ENVS

    @dd.virtualfield(dd.HtmlBox(_("Result")))
    def result(self,ar):
        return self.response_xml
        
#~ dd.update_field(CBSSRequest,'project',blank=False,null=False)
dd.update_field(CBSSRequest,'user',blank=False,null=False)

class CBSSRequestDetail(dd.FormLayout):
    #~ main = 'request response'
    main = 'request technical'
    
    request = """
    info
    parameters
    result
    """
    
    technical = """
    response_xml 
    info_messages
    debug_messages
    """
    
    info = """
    id person user environment sent status ticket
    """
    
    
    #~ response = "response_xml\nlogged_messages"
    
    def setup_handle(self,lh):
        lh.request.label = _("Request")
        lh.info.label = _("Request information")
        lh.result.label = _("Result")
        lh.technical.label = _("Technical")
        #~ lh.technical.required_user_level = UserLevels.manager
        #~ lh.technical.read_permission = perms.Required(user_groups = ['cbss'])
        #~ lh.technical.required = dict(user_level='admin')
        lh.technical.required.update(user_level='admin')

        #~ lh.response.label = _("Response")
        #~ lh.log.label = _("Log")
        lh.parameters.label = _("Parameters")
    

class SSDNRequest(CBSSRequest):
    """
    Abstract Base Class for Models that represent 
    SSDN ("classic") requests to the :term:`CBSS`.
    """
    
    wsdl_parts = ('cache','wsdl','WebServiceConnector.wsdl')
    
    xsd_filename = None
    
    class Meta:
        abstract = True
        
    def validate_against_xsd(self,srvreq,xsd_filename):
        #~ logger.info("20120524 Validate against %s", xsd_filename)
        from lxml import etree
        xml = unicode(srvreq)
        #~ print xml
        doc = etree.fromstring(xml)
        schema_doc = etree.parse(xsd_filename)
        schema = etree.XMLSchema(schema_doc)
        #~ if not schema.validate(doc):
            #~ print xml
        schema.assertValid(doc)
        #~ self.logmsg("Validated %s against %s", xml,xsd_filename)
        self.logmsg_debug("Validated %s against %s", self,xsd_filename)
      
    def validate_wrapped(self,srvreq):
        self.validate_against_xsd(srvreq,xsdpath('SSDN','Service','SSDNRequest.xsd'))
        
        
    def validate_inner(self,srvreq):
        if not self.xsd_filename: return
        self.validate_against_xsd(srvreq,self.xsd_filename)
        
    
    def validate_request(self):
        """
        Validates the generated XML against the XSD files.
        Used by test suite.
        It is not necessary to validate each real request before actually sending it.
        """
        srvreq = self.build_request()
        self.validate_inner(srvreq)
        wrapped_srvreq = self.wrap_ssdn_request(srvreq,datetime.datetime.now())
        self.validate_wrapped(wrapped_srvreq)
        self.logmsg_info(_("Request has been validated against XSD files"))
      
    def create_client(self):
        url = self.get_wsdl_uri()
        #~ logger.info("Instantiate Client at %s", url)
        t = HttpTransport()
        client = Client(url, transport=t)
        #~ print 20120507, client
        return client
        
    def execute_request_(self,now,simulate_response):
        """
        SSDN specific part of a request.
        """
        srvreq = self.build_request()
        
        #~ if validate:
            #~ self.validate_inner(srvreq)
            
        wrapped_srvreq = self.wrap_ssdn_request(srvreq,now)
        
        #~ if validate:
            #~ self.validate_wrapped(wrapped_srvreq)
            #~ logger.info("XSD validation passed.")
            
        if simulate_response is None:
            # this is the normal case
            self.check_environment(srvreq)
            
            client = get_client(self)
        
            xmlString = unicode(wrapped_srvreq)
            self.request_xml = xmlString
            #~ logger.info("20120521 Gonna sendXML(<xmlString>):\n%s",xmlString)
            if not settings.LINO.cbss_live_tests:
                #~ raise Warning("NOT sending because `cbss_live_tests` is False:\n" + unicode(xmlString))
                raise Warning(
                    "NOT sending because `cbss_live_tests` is False:\n" + xmlString)
            #~ xmlString.append(wrapped_srvreq)
            self.logmsg_debug("client.service.sendXML(\n%s\n)",xmlString)
            res = client.service.sendXML(xmlString)
            #~ print 20120522, res
            self.response_xml = unicode(res)
            #~ self.save()
            #~ return self.fill_from_string(res.encode('utf-8'),xmlString)
            return self.fill_from_string(res.encode('utf-8'))
        else:
            self.environment = 'demo'
            self.response_xml = unicode(simulate_response)
            return self.fill_from_string(simulate_response)
                
        
    def fill_from_string(self,s,sent_xmlString=None):
        #~ self.response_xml = unicode(res)
        reply = PARSER.parse(string=s).root()
        self.ticket = nodetext(reply.childAtPath('/ReplyContext/Message/Ticket'))
        rs = reply.childAtPath('/ServiceReply/ResultSummary')
        if rs is None:
            raise Warning("Missing ResultSummary in :\n%s" % reply)
            
        for dtl in rs.getChildren('Detail'):
        #~ for detail in rs.getChildren():
            msg = nodetext(dtl.childAtPath('/Severity')) # WARNING, INFO, ERROR...
            msg += " " + nodetext(dtl.childAtPath('/ReasonCode'))
            msg += " (%s) : " % nodetext(dtl.childAtPath('/AuthorCodeList'))
            msg += nodetext(dtl.childAtPath('/Diagnostic'))
            #~ print '========'
            #~ print msg
            #~ raise Warning(msg)
            self.logmsg_info(msg)
            
        rc = nodetext(rs.childAtPath('/ReturnCode'))
        #~ print reply.__class__, dir(reply)
        #~ print reply
        #~ rc = reply.root().SSDNReply.ServiceReply.ResultSummary.ReturnCode
        
        if rc == '0':
            self.status = RequestStates.ok
        elif rc == '1':
            self.status = RequestStates.warnings
            #~ self.logmsg_debug("Warnings:==============\n%s\n===============" % s)
        #~ elif rc == '10000':
            #~ self.status = RequestStates.errors
        else:
            self.status = RequestStates.errors
            #~ self.response_xml = unicode(reply)
            #~ dtl = rs.childAtPath('/Detail')
            #~ msg = CBSS_ERROR_MESSAGE % rc
            #~ keys = ('Severity', 'ReasonCode', 'Diagnostic', 'AuthorCodeList')
            #~ msg += '\n'.join([
                #~ k+' : '+nodetext(dtl.childAtPath('/'+k))
                    #~ for k in keys])
            #~ raise Warning(msg)
            #~ return None
            #~ raise Exception("Got invalid response status")
            
        #~ self.on_cbss_ok(reply)
        service_reply = self.get_service_reply(reply)
        if service_reply is None:
            raise Warning("Got response without service reply.")
            #~ raise Exception(
              #~ "Return code is %r, but there's no service reply." % rc)
              #~ "Return code is %r but there's no service reply in:\n%s\n" % (rc,reply))
        #~ reply.childAtPath('/ServiceReply/IdentifyPersonReply')
        self.response_xml = unicode(service_reply)
        return service_reply
        
        
    def get_service_reply(self,full_reply=None):
        raise NotImplementedError()
        
        
        
    def wrap_ssdn_request(self,srvreq,dt):
        """
        Wrap the given service request into the SSDN envelope 
        by adding AuthorizedUser and other information common 
        the all SSDN requests).
        """
        #~ up  = settings.LINO.ssdn_user_params
        #~ user_params = settings.LINO.cbss_user_params
        sc = settings.LINO.site_config
        #~ au = E('common:AuthorizedUser',ns=NSCOMMON)
        #~ au.append(E('common:UserID').setText(up['UserID']))
        #~ au.append(E('common:Email').setText(up['Email']))
        #~ au.append(E('common:OrgUnit').setText(up['OrgUnit']))
        #~ au.append(E('common:MatrixID').setText(up['MatrixID']))
        #~ au.append(E('common:MatrixSubID').setText(up['MatrixSubID']))
        au = E('ssdn:AuthorizedUser')
        #~ au.append(E('ssdn:UserID').setText(user_params['UserID']))
        au.append(E('ssdn:UserID').setText(sc.ssdn_user_id))
        #~ au.append(E('ssdn:Email').setText(user_params['Email']))
        #~ if not sc.site_company:
            #~ raise Exception("")
        #~ au.append(E('ssdn:Email').setText(sc.site_company.email))
        au.append(E('ssdn:Email').setText(sc.ssdn_email))
        #~ au.append(E('ssdn:OrgUnit').setText(user_params['OrgUnit']))
        #~ au.append(E('ssdn:OrgUnit').setText(sc.site_company.vat_id))
        au.append(E('ssdn:OrgUnit').setText(sc.cbss_org_unit))
        #~ au.append(E('ssdn:MatrixID').setText(user_params['MatrixID']))
        au.append(E('ssdn:MatrixID').setText(sc.sector.code))
        #~ au.append(E('ssdn:MatrixSubID').setText(user_params['MatrixSubID']))
        au.append(E('ssdn:MatrixSubID').setText(sc.sector.subcode))
        
        ref = "%s # %s" % (self.__class__.__name__,self.id)
        msg = E('ssdn:Message')
        msg.append(E('ssdn:Reference').setText(ref))
        msg.append(E('ssdn:TimeRequest').setText(dt.strftime("%Y%m%dT%H%M%S")))
        
        context = E('ssdn:RequestContext')
        context.append(au)
        context.append(msg)
        
        sr = E('ssdn:ServiceRequest')
        sr.append(E('ssdn:ServiceId').setText(self.ssdn_service_id))
        sr.append(E('ssdn:Version').setText(self.ssdn_service_version))
        sr.append(srvreq)
        
        
        #~ xg.set_default_namespace(SSDN)
        e = E('ssdn:SSDNRequest',ns=NSSSDN)
        e.append(context)
        e.append(sr)
        #~ if srvreq.prefix != e.prefix:
            #~ e.addPrefix(srvreq.prefix,srvreq.nsprefixes[srvreq.prefix])
        
        return e
      


class NewStyleRequest(CBSSRequest):
    """
    Abstract Base Class for Models that represent 
    "new style" requests to the :term:`CBSS` (and responses).
    """
    
    class Meta:
        abstract = True
        
    def create_client(self):
        url = self.get_wsdl_uri()
        
        #~ logger.info("Instantiate Client at %s", url)
        sc = settings.LINO.site_config
        #~ t = HttpAuthenticated(
            #~ username=settings.LINO.cbss_username, 
            #~ password=settings.LINO.cbss_password)
        t = HttpAuthenticated(
            username=sc.cbss_http_username, 
            password=sc.cbss_http_password)
        client = Client(url, transport=t,retxml=True)
        #~ print 20120613, client
        return client
        
    def execute_request_(self,now,simulate_response):
      
        client = get_client(self)

        #~ sc = settings.LINO.site_config
        ci = client.factory.create('ns0:CustomerIdentificationType')
        #~ cbeNumber = client.factory.create('ns0:CbeNumberType')
        #~ ci.cbeNumber = settings.LINO.cbss_cbe_number
        #~ ci.cbeNumber = settings.LINO.site_config.site_company.vat_id
        ci.cbeNumber = settings.LINO.site_config.cbss_org_unit
        info = client.factory.create('ns0:InformationCustomerType')
        info.ticket = str(self.id)
        info.timestampSent = now
        info.customerIdentification = ci
        
        return self.execute_newstyle(client,info,simulate_response)
        
        
        
    def on_cbss_ok(self,reply):
        """
        Called when a successful reply has been received.
        """
        pass
        
    #~ def __unicode__(self):
        #~ return u"%s#%s" % (self.__class__.__name__,self.pk)
        
    def get_service_reply(self):
        #~ """
    #~ Example of a reply::

    #~ (reply){
       #~ informationCustomer =
          #~ (InformationCustomerType){
             #~ ticket = "1"
             #~ timestampSent = 2012-05-23 09:24:55.316312
             #~ customerIdentification =
                #~ (CustomerIdentificationType){
                   #~ cbeNumber = "0123456789"
                #~ }
          #~ }
       #~ informationCBSS =
          #~ (InformationCBSSType){
             #~ ticketCBSS = "f11736b3-97bc-452a-a75c-16fcc2a2f6ae"
             #~ timestampReceive = 2012-05-23 08:24:37.000385
             #~ timestampReply = 2012-05-23 08:24:37.000516
          #~ }
       #~ status =
          #~ (StatusType){
             #~ value = "NO_RESULT"
             #~ code = "MSG00008"
             #~ description = "A validation error occurred."
             #~ information[] =
                #~ (InformationType){
                   #~ fieldName = "ssin"
                   #~ fieldValue = "12345678901"
                #~ },
          #~ }
       #~ searchInformation =
          #~ (SearchInformationType){
             #~ ssin = "12345678901"
             #~ language = "de"
             #~ history = False
          #~ }
     #~ }        
        #~ """
        if not self.response_xml: return None
        client = get_client(self).service
        #~ print '20120613b', dir(client)
        return client.succeeded(client.method.binding.input, self.response_xml)
        
        
    def execute_newstyle(self,client,infoCustomer,simulate_response):
        raise NotImplementedError()





class SSIN(dd.Model):
    """
    Abstract base for Requests that have a field `national_id` and a method 
    :meth:`get_ssin`.
    """
    class Meta:
        abstract = True
  
    national_id = models.CharField(max_length=200,
        blank=True,verbose_name=_("National ID")
        ,validators=[pcsw.niss_validator]
        )
        
    def get_ssin(self):
        national_id = self.national_id.replace('=','')
        national_id = national_id.replace(' ','')
        national_id = national_id.replace('-','')
        return national_id
        
    #~ def save(self,*args,**kw):
        #~ if self.person_id and not self.last_name: 
            #~ self.fill_from_person(self.person)
        #~ super(SSIN,self).save(*args,**kw)
        
    def on_create(self,ar):
        #~ print '20120629 SSIN.on_create', obj2str(self), ar
        #~ super(ContractBase,self).on_create(request)
        self.person_changed(ar)
        super(SSIN,self).on_create(ar)
        
    def person_changed(self,ar):
        #~ raise Exception("20120704")
        #~ print '20120704 person_changed'
        if self.person_id: 
            self.fill_from_person(self.person)
        
    def fill_from_person(self,person):
        self.national_id = person.national_id
        
        
class WithPerson(SSIN):
    """
    Mixin for models that have certain fields
    """
    class Meta:
        abstract = True
  
    birth_date = dd.IncompleteDateField(
        blank=True,
        verbose_name=_("Birth date"))
        
    sis_card_no = models.CharField(verbose_name=_('SIS card number'),
        max_length=10,
        blank=True, help_text="""\
The number of the SIS card used to authenticate the person.""")

    id_card_no = models.CharField(verbose_name=_('ID card number'),
        max_length=20,
        blank=True, help_text="""\
The number of the ID card used to authenticate the person.""")

    first_name = models.CharField(max_length=200,
      blank=True,
      verbose_name=_('First name'))
    "Space-separated list of all first names."
    
    last_name = models.CharField(max_length=200,
      blank=True,
      verbose_name=_('Last name'))
    """Last name (family name)."""
    
        
    def fill_from_person(self,person):
        self.national_id = person.national_id
        self.id_card_no = person.card_number
        self.last_name = person.last_name
        self.first_name = person.first_name
        self.birth_date = person.birth_date
        #~ print '20120603 fill_from_person', self.national_id
                
        
class IdentifyPersonRequest(SSDNRequest,WithPerson):
    """
    A request to the IdentifyPerson service.
    
    """
    
    ssdn_service_id = 'OCMWCPASIdentifyPerson'
    ssdn_service_version = '20050930'
    xsd_filename = xsdpath('SSDN','OCMW_CPAS',
        'IdentifyPerson','IdentifyPersonRequest.xsd')
        #~ 'IDENTIFYPERSON','IDENTIFYPERSONREQUEST.XSD')
    
    #~ cbss_namespace = cbss.IPR # IdentifyPersonRequest
    
    class Meta:
        verbose_name = _("IdentifyPerson Request")
        verbose_name_plural = _("IdentifyPerson Requests")
        
    #~ first_name = models.CharField(max_length=200,
      #~ blank=True,
      #~ verbose_name=_('First name'))
    #~ "Space-separated list of all first names."
    
    #~ last_name = models.CharField(max_length=200,
      #~ blank=True,
      #~ verbose_name=_('Last name'))
    #~ """Last name (family name)."""

    middle_name = models.CharField(max_length=200,
      blank=True,
      verbose_name=_('Middle name'),
      help_text="Whatever this means...")
      
    gender = Gender.field()
        
    tolerance = models.IntegerField(verbose_name=_('Tolerance'),
      default=0,
      help_text=u"""
      Falls Monat oder Tag des Geburtsdatums unbekannt sind, 
      um wieviel Monate bzw. Tage die Suche nach unten/oben ausgeweitet wird.
      Gültige Werte: 0 bis 10.
      """)
      # 20120606 gridcolumn doesn't like tooltips containing HTML
      #~ <p>Zum Beispiel 
      #~ <table border=1 class="htmlText">
      #~ <tr>
        #~ <td>Geburtsdatum</td>
        #~ <td colspan="3">Toleranz</td>
      #~ </tr><tr>
        #~ <td></td>
        #~ <td>0</td>
        #~ <td>1</td>
        #~ <td>10</td>
      #~ </tr><tr>
        #~ <td> 1968-00-00  </td>
        #~ <td> im Jahr 1968 </td>
        #~ <td> von 1967 bis 1969 </td>
        #~ <td> 1958 bis 1978 </td>
      #~ </tr><tr>
        #~ <td> 1968-06-00  </td>
        #~ <td> im Juni 1968 </td>
        #~ <td> von Mai  bis Juli 1968 </td>
        #~ <td>von Oktober 1967 bis April 1969</td>
      #~ </tr>
      #~ </table>
      #~ </p>
      
      
        
    #~ def on_create(self,ar):
        #~ mixins.AutoUser.on_create(self,ar)
        #~ SSIN.on_create(self,ar)

    def get_result_table(self,ar):
        return ar.spawn(IdentifyPersonResult,master_instance=self)
        
    def fill_from_person(self,person):
        self.national_id = person.national_id
        self.id_card_no = person.card_number
        self.birth_date = person.birth_date
        if not self.national_id:
            self.gender = person.gender
            self.last_name = person.last_name
            self.first_name = person.first_name
        
    def build_request(self):
        """Construct and return the root element of the (inner) service request."""
        #~ if not self.birth_date:
            #~ raise Warning("Empty birth date (a full_clean() would have told that, too!)")
            #~ raise Warning(_("Birth date may not be empty."))
        
        national_id = self.get_ssin()
        gender = gender2cbss(self.gender)
        #~ https://fedorahosted.org/suds/wiki/TipsAndTricks#IncludingLiteralXML
        main = E('ipr:IdentifyPersonRequest',ns=NSIPR)
        sc = E('ipr:SearchCriteria') 
        main.append(sc)
        if national_id:
            # VerificatioinData is ignored if there's no SSIN in the SearchCriteria
            sc.append(E('ipr:SSIN').setText(national_id))
            
            vd = E('ipr:VerificationData')
            main.append(vd)
            if self.sis_card_no:
                vd.append(E('ipr:SISCardNumber').setText(self.sis_card_no))
            if self.id_card_no:
                vd.append(E('ipr:IdentityCardNumber').setText(self.id_card_no))
            
            pd = E('ipr:PersonData')
            vd.append(pd)
            #~ if not self.last_name or not self.first_name:
                #~ raise Warning("Fields last_name and first_name are mandatory.")
            pd.append(E('ipr:LastName').setText(self.last_name))
            pd.append(E('ipr:FirstName').setText(self.first_name))
            pd.append(E('ipr:MiddleName').setText(self.middle_name))
            pd.append(E('ipr:BirthDate').setText(str(self.birth_date)))
            #~ if not self.birth_date.is_complete():
                #~ pd.append(E('ipr:Tolerance').setText(self.tolerance))
            #~ if gender is not None: pd.append(E('ipr:Gender').setText(gender))
        pc = E('ipr:PhoneticCriteria') 
        sc.append(pc)
        pc.append(E('ipr:LastName').setText(self.last_name))
        pc.append(E('ipr:FirstName').setText(self.first_name))
        pc.append(E('ipr:MiddleName').setText(self.middle_name))
        pc.append(E('ipr:BirthDate').setText(str(self.birth_date)))
        return main
        
    def get_service_reply(self,full_reply=None):
        if full_reply is not None:
            return full_reply.childAtPath('/ServiceReply/IdentifyPersonReply')
        return PARSER.parse(string=self.response_xml.encode('utf-8')).root()
        #~ return reply
           
        #~ if False:
          
            #~ try:
                #~ res = self.cbss_namespace.execute(srvreq,str(self.id),now)
            #~ except cbss.Warning,e:
                #~ self.status = RequestStates.exception
                #~ self.response_xml = unicode(e)
                #~ self.save()
                #~ return
            #~ except Exception,e:
                #~ self.status = RequestStates.exception
                #~ self.response_xml = traceback.format_exc(e)
                #~ self.save()
                #~ return
            #~ self.sent = now
            #~ self.response_xml = res.data.xmlString
            #~ reply = cbss.xml2reply(res.data.xmlString)
            #~ rc = reply.ServiceReply.ResultSummary.ReturnCode
            #~ if rc == '0':
                #~ self.status = RequestStates.ok
            #~ elif rc == '1':
                #~ self.status = RequestStates.warnings
            #~ elif rc == '10000':
                #~ self.status = RequestStates.errors
            #~ self.save()
            
            #~ if self.status != RequestStates.ok:
                #~ msg = '\n'.join(list(cbss.reply2lines(reply)))
                #~ raise Exception(msg)
                
            #~ self.on_cbss_ok(reply)
        

dd.update_field(IdentifyPersonRequest,'birth_date',blank=False)
"""
DocumentInvalid
Element '{http://www.ksz-bcss.fgov.be/XSD/SSDN/OCMW_CPAS/IdentifyPerson}BirthDate': [facet 'length'] The value has a length of '0'; this differs from the allowed length of '10'., line 7

"""
#~ dd.update_field(IdentifyPersonRequest,'first_name',blank=True)
#~ dd.update_field(IdentifyPersonRequest,'last_name',blank=True)

    
    
class IdentifyPersonRequestDetail(CBSSRequestDetail):
    p1 = """
    national_id
    spacer
    """
    p2 = """
    first_name middle_name last_name
    birth_date tolerance  gender 
    """
    parameters = "p1 p2"
    
    result = "IdentifyPersonResult"
    
    def setup_handle(self,lh):
        lh.p1.label = _("Using the national ID")
        lh.p2.label = _("Using phonetic search")
        CBSSRequestDetail.setup_handle(self,lh)
    
class IdentifyPersonRequestInsert(IdentifyPersonRequestDetail):
    window_size = (60,'auto')
    main = """
    person national_id
    p2
    """
    p2 = """
    first_name middle_name last_name
    birth_date tolerance  gender 
    """
    
    def setup_handle(self,lh):
        lh.p2.label = _("Phonetic search")

class CBSSRequests(dd.Table):
    #~ create_required = dict(user_level='user')
    pass
    #~ @classmethod
    #~ def get_row_permission(cls,action,user,row):
        #~ if row.ticket and not action.readonly: 
            #~ return False
        #~ if not super(CBSSRequests,cls).get_row_permission(action,user,row):
            #~ return False
        #~ return True
        
    
class IdentifyPersonRequests(CBSSRequests):
    #~ window_size = (500,400)
    required = dict(user_groups='cbss')
    #~ required_user_groups = ['cbss']
    model = IdentifyPersonRequest
    active_fields = ['person']
    detail_layout = IdentifyPersonRequestDetail()
    insert_layout = IdentifyPersonRequestInsert()
    
    @dd.constant()
    def spacer(self,ui):  return '<br/>'
    
class MyIdentifyPersonRequests(mixins.ByUser,IdentifyPersonRequests):
    pass
    
class IdentifyRequestsByPerson(IdentifyPersonRequests):
    master_key = 'person'
    column_names = 'user sent status *'
    



class IdentifyPersonResult(dd.VirtualTable):
    """
    Displays the response of an :class:`IdentifyPersonRequest`
    as a table.
    """
    master = IdentifyPersonRequest
    master_key = None
    label = _("Results")
    column_names = 'national_id:10 last_name:20 first_name:10 address birth_date:10 birth_location civil_state *'
    
    @classmethod
    def get_data_rows(self,ar):
        ipr = ar.master_instance
        if ipr is None: 
            #~ print "20120606 ipr is None"
            return
        #~ if not ipr.status in (RequestStates.ok,RequestStates.fictive):
        if not ipr.status in (RequestStates.ok,RequestStates.warnings):
            #~ print "20120606 wrong status", ipr.status 
            return
        service_reply = ipr.get_service_reply()
        results = service_reply.childAtPath('/SearchResults').children
        #~ print "20120606 got", service_reply
        if results is None:
            #~ print "20120606 no /SearchResults"
            #~ return []
            return
        for obj in results:
            data = dict()
            data.update(national_id = nodetext(obj.childAtPath('/Basic/SocialSecurityUser')))
            data.update(last_name = nodetext(obj.childAtPath('/Basic/LastName')))
            data.update(first_name = nodetext(obj.childAtPath('/Basic/FirstName')))
            data.update(gender = cbss2gender(nodetext(obj.childAtPath('/Basic/Gender'))))
            data.update(birth_date = cbss2date(nodetext(obj.childAtPath('/Basic/BirthDate'))))
            data.update(civil_state = cbss2civilstate(
                obj.childAtPath('/Extended/CivilState')))
            data.update(birth_location=nodetext(obj.childAtPath('/Extended/BirthLocation')))
            data.update(cbss2address(obj))
            yield AttrDict(**data)
        #~ return results
        #~ if service_reply is not None:
            #~ results = service_reply.childAtPath('/SearchResults')
            #~ if results is not None:
            
    @dd.displayfield(_("National ID"))
    def national_id(self,obj,ar):
        #~ return obj.childAtPath('/Basic/SocialSecurityUser').text
        return obj.national_id
            
    @dd.displayfield(_("Last name"))
    def last_name(self,obj,ar):
        return obj.last_name
        #~ return obj.childAtPath('/Basic/LastName').text
        
    @dd.displayfield(_("First name"))
    def first_name(self,obj,ar):
        return obj.first_name
        #~ return obj.childAtPath('/Basic/FirstName').text
            
    @dd.virtualfield(Gender.field())
    def gender(self,obj,ar):
        return obj.gender
        #~ return cbss2gender(obj.childAtPath('/Basic/Gender').text)
            
    #~ @dd.displayfield(dd.IncompleteDateField(_("Birth date")))
    @dd.displayfield(_("Birth date"))
    def birth_date(self,obj,ar):
        return obj.birth_date
        #~ return obj.childAtPath('/Basic/BirthDate').text
        
    @dd.displayfield(_("Birth location"))
    def birth_location(self,obj,ar):
        return obj.birth_location
        
    @dd.displayfield(_("Civil state"))
    def civil_state(self,obj,ar):
        return obj.civil_state
        #~ return obj.childAtPath('/Extended/CivilState').text
        
    @dd.displayfield(_("Address"))
    def address(self,obj,ar):
        return obj.address
        
        
            
    #~ @dd.virtualfield(models.ForeignKey(settings.LINO.person_model))
    #~ @dd.displayfield(_("Person"))
    #~ def person(self,obj,ar):
        #~ from lino.apps.pcsw.models import Person
        #~ niss = obj.childAtPath('/Basic/SocialSecurityUser').text
        #~ if niss:
            #~ try:
                #~ return unicode(Person.objects.get(national_id=niss))
            #~ except Person.DoesNotExist:
                #~ pass
        #~ return ''
            






class ManageAction(ChoiceList):
    """
    Possible values for the 
    `action` field of a :class:`ManageAccessRequest`.
    """
    label = _("Action")

add = ManageAction.add_item
add('1',_("Register"),'REGISTER')
add('2',_("Unregister"),'UNREGISTER')
add('3',_("List"),'LIST')


class QueryRegister(ChoiceList):
    """
    Possible values for the 
    `query_register` field of a :class:`ManageAccessRequest`.
    
    """
    label = _("Query Register")

add = QueryRegister.add_item
add('1',_("Primary"),'PRIMARY')
add('2',_("Secondary"),'SECONDARY')
add('3',_("All"),'ALL')

class ManageAccessRequest(SSDNRequest,WithPerson):
    """
    A request to the ManageAccess service.
    
    Registering a person means that this PCSW is 
    going to maintain a dossier about this person.
    Users commonly say "to integrate" a person.
    
    """
  
    ssdn_service_id = 'OCMWCPASManageAccess'
    ssdn_service_version = '20050930'
    
    xsd_filename = xsdpath('SSDN','OCMW_CPAS',
        'ManageAccess','ManageAccessRequest.xsd')
    
    class Meta:
        verbose_name = _("ManageAccess Request")
        verbose_name_plural = _("ManageAccess Requests")
        
    #~ purpose = models.IntegerField(verbose_name=_('Purpose'),
      #~ default=0,help_text="""\
#~ The purpose for which the inscription needs to be 
#~ registered/unregistered or listed. 
#~ For listing this field is optional, 
#~ for register/unregister it is obligated.""")

    #~ sector = models.IntegerField(verbose_name=_('Sector'),
      #~ blank=False,default=0,help_text="""\
#~ For register and unregister this element is ignored. 
#~ It can be used for list, 
#~ when information about sectors is required.""")

    sector = models.ForeignKey(Sector,
      #~ blank=True, 
      editable=False,
      help_text="""\
For register and unregister this element is ignored. 
It can be used for list, 
when information about sectors is required.""")

    purpose = models.ForeignKey(Purpose,
      #~ blank=True,null=True,
      help_text="""\
The purpose for which the inscription needs to be 
registered/unregistered or listed. 
For listing this field is optional, 
for register/unregister it is mandatory.""")

    start_date = models.DateField(
        #~ blank=True,null=True,
        verbose_name=_("Period from"))
    end_date = models.DateField(
        #~ blank=True,null=True,
        verbose_name=_("Period until"))

    # 20120527 : Django converts default value to unicode. didnt yet understand why.
    action = ManageAction.field(blank=False,default=ManageAction.LIST)
    query_register = QueryRegister.field(blank=False,default=QueryRegister.ALL)
    #~ action = ManageAction.field(blank=False)
    #~ query_register = QueryRegister.field(blank=False) # ,default=QueryRegister.ALL)

    def save(self,*args,**kw):
        if not self.sector_id:
            self.sector = settings.LINO.site_config.sector
        super(ManageAccessRequest,self).save(*args,**kw)

    @chooser()
    def purpose_choices(cls,sector):
        if not sector:
            sector = settings.LINO.site_config.sector
        if not sector:
            raise Exception("SiteConfig.sector is not set!")
        Q = models.Q
        return Purpose.objects.filter(
            Q(sector_code=sector.code)|Q(sector_code__isnull=True)).order_by('code')


    #~ def on_create(self,ar):
        #~ mixins.AutoUser.on_create(self,ar)
        #~ SSIN.on_create(self,ar)

    def build_request(self):
        """Construct and return the root element of the (inner) service request."""
        national_id = self.get_ssin()
        main = E('mar:ManageAccessRequest',ns=NSMAR)
        main.append(E('mar:SSIN').setText(national_id))
        #~ main.append(E('mar:Purpose').setText(str(self.purpose)))
        if self.purpose_id:
            main.append(E('mar:Purpose').setText(str(self.purpose.code)))
        period = E('mar:Period')
        main.append(period)
        period.append(E('common:StartDate',ns=NSCOMMON).setText(str(self.start_date)))
        if self.end_date:
            period.append(E('common:EndDate',ns=NSCOMMON).setText(str(self.end_date)))
        main.append(E('mar:Action').setText(self.action.name))
        main.append(E('mar:Sector').setText(str(self.sector.code)))
        if self.query_register:
            main.append(E('mar:QueryRegister').setText(self.query_register.name))
        proof = E('mar:ProofOfAuthentication')
        main.append(proof)
        if self.sis_card_no:
            proof.append(E('mar:SISCardNumber').setText(self.sis_card_no))
        if self.id_card_no:
            proof.append(E('mar:IdentityCardNumber').setText(self.id_card_no))
        if self.last_name or self.first_name or self.birth_date:
            pd = E('mar:PersonData')
            proof.append(pd)
            pd.append(E('mar:LastName').setText(self.last_name))
            pd.append(E('mar:FirstName').setText(self.first_name))
            pd.append(E('mar:BirthDate').setText(self.birth_date))
        return main
    
    def get_service_reply(self,full_reply=None):
        """
        Extract the "service reply" part from a full reply.
        Example of a full reply::
        
         <ServiceReply>
            <ns2:ResultSummary xmlns:ns2="http://www.ksz-bcss.fgov.be/XSD/SSDN/Common" ok="YES">
               <ns2:ReturnCode>0</ns2:ReturnCode>
            </ns2:ResultSummary>
            <ServiceId>OCMWCPASManageAccess</ServiceId>
            <Version>20050930</Version>
            <ns3:ManageAccessReply xmlns:ns3="http://www.ksz-bcss.fgov.be/XSD/SSDN/OCMW_CPAS/ManageAccess">
               <ns3:OriginalRequest>
                  <ns3:SSIN>68060105329</ns3:SSIN>
                  <ns3:Purpose>1</ns3:Purpose>
                  <ns3:Period>
                     <ns4:StartDate xmlns:ns4="http://www.ksz-bcss.fgov.be/XSD/SSDN/Common">2012-05-24+02:00</ns4:StartDate>
                     <ns5:EndDate xmlns:ns5="http://www.ksz-bcss.fgov.be/XSD/SSDN/Common">2012-05-24+02:00</ns5:EndDate>
                  </ns3:Period>
                  <ns3:Action>REGISTER</ns3:Action>
               </ns3:OriginalRequest>
               <ns3:Registrations>
                  <ns3:Purpose>1</ns3:Purpose>
                  <ns3:Period>
                     <ns6:StartDate xmlns:ns6="http://www.ksz-bcss.fgov.be/XSD/SSDN/Common">2012-05-24+02:00</ns6:StartDate>
                     <ns7:EndDate xmlns:ns7="http://www.ksz-bcss.fgov.be/XSD/SSDN/Common">2012-05-24+02:00</ns7:EndDate>
                  </ns3:Period>
                  <ns3:OrgUnit>63023</ns3:OrgUnit>
                  <ns3:Register>SECONDARY</ns3:Register>
               </ns3:Registrations>
            </ns3:ManageAccessReply>
         </ServiceReply>        
        """
        if full_reply is not None:
            return full_reply.childAtPath('/ServiceReply/ManageAccessReply')
        return PARSER.parse(string=self.response_xml.encode('utf-8')).root()
        
        
    
dd.update_field(ManageAccessRequest,'national_id',blank=False,help_text="""\
The SSIN of the person to register/unregister/list.
""")





class ManageAccessRequestDetail(CBSSRequestDetail):
    p1 = """
    action start_date end_date 
    purpose query_register
    """
    proof = """
    national_id sis_card_no id_card_no
    first_name last_name birth_date 
    """
    parameters = "p1 proof"
    
    def setup_handle(self,lh):
        lh.p1.label = _("Requested action")
        lh.proof.label = _("Proof of authentication")
        CBSSRequestDetail.setup_handle(self,lh)
        
class ManageAccessRequestInsert(dd.FormLayout):
    window_size = (60,'auto')
    
    p1 = """
    action start_date end_date 
    purpose query_register
    """
    proof = """
    national_id sis_card_no id_card_no
    first_name last_name birth_date 
    """
    main = """
    person
    p1 
    proof
    """
    
    def setup_handle(self,lh):
        lh.p1.label = _("Requested action")
        lh.proof.label = _("Proof of authentication")
        super(ManageAccessRequestInsert,self).setup_handle(lh)
    
        
    

class ManageAccessRequests(CBSSRequests):
    #~ window_size = (500,400)
    model = ManageAccessRequest
    detail_layout = ManageAccessRequestDetail()
    insert_layout = ManageAccessRequestInsert()
    required = dict(user_groups='cbss')
    #~ required_user_groups = ['cbss']
    active_fields = ['person']
    
    
class ManageAccessRequestsByPerson(ManageAccessRequests):
    master_key = 'person'


class MyManageAccessRequests(ManageAccessRequests,mixins.ByUser):
    pass


from lino.modlib.cbss.tx25 import *
    

from lino.models import SiteConfig
dd.inject_field(SiteConfig,
    'sector',
    models.ForeignKey(Sector,
        blank=True,null=True,
        help_text="""\
The CBSS sector/subsector of the requesting organization.        
For PCSWs this is always 17.1.
Used in SSDN requests as text of the `MatrixID` and `MatrixSubID` 
elements of `AuthorizedUser`. 
Used in ManageAccess requests as default value 
for the non-editable field `sector` 
(which defines the choices of the `purpose` field).
"""))
    
dd.inject_field(SiteConfig,
    'cbss_org_unit',
    models.CharField(_("Requesting organisation"),
      max_length=50,
      blank=True,
      help_text="""\
In CBSS requests, identifies the requesting organization.
For PCSWs this is the enterprise number 
(CBE, KBO) and should have 10 digits and no formatting characters.

Used in SSDN requests as text of the `AuthorizedUser\OrgUnit` element . 
Used in new style requests as text of the `CustomerIdentification\cbeNumber` element . 
"""))
dd.inject_field(SiteConfig,
    'ssdn_user_id',
    models.CharField(_("SSDN User Id"),
      max_length=50,
      blank=True,
      help_text="""\
Used in SSDN requests as text of the `AuthorizedUser\UserID` element.
"""))
dd.inject_field(SiteConfig,
    'ssdn_email',
    models.EmailField(_("SSDN email address"),
      blank=True,
      help_text="""\
Used in SSDN requests as text of the `AuthorizedUser\Email` element.
"""))
dd.inject_field(SiteConfig,
    'cbss_http_username',
    models.CharField(_("HTTP username"),
      max_length=50,
      blank=True,
      help_text="""\
Used in the http header of new-style requests.
"""))
dd.inject_field(SiteConfig,
    'cbss_http_password',
    models.CharField(_("HTTP password"),
      max_length=50,
      blank=True,
      help_text="""\
Used in the http header of new-style requests.
"""))


    
    
dd.inject_quick_add_buttons(
    pcsw.Person,'cbss_identify_person',IdentifyRequestsByPerson)
dd.inject_quick_add_buttons(
    pcsw.Person,'cbss_manage_access',ManageAccessRequestsByPerson)
dd.inject_quick_add_buttons(
    pcsw.Person,'cbss_retrieve_ti_groups',RetrieveTIGroupsRequestsByPerson)

def cbss_summary(self,ar):
    """
    returns a summary overview of the CBSS requests for this person.
    """
    #~ qs = IdentifyPersonRequest.objects.filter(person=self,status=RequestStates.ok)
    html = '<p><ul>'
    #~ for m in (IdentifyPersonRequest,ManageAccessRequest,RetrieveTIGroupsRequest):
        #~ n = m.objects.filter(person=self).count()
        #~ if n > 0:
            #~ html += "<li>%d %s</li>" % (n,unicode(m._meta.verbose_name_plural))
    #~ html += '</ul></p>'
    #~ html += '<p>Using XyzByPerson:<ul>'
    for t in (IdentifyRequestsByPerson,ManageAccessRequestsByPerson,RetrieveTIGroupsRequestsByPerson):
        n = ar.spawn(t,master_instance=self).get_total_count()
        if n > 0:
            html += "<li>%d %s</li>" % (n,unicode(t.model._meta.verbose_name_plural))
    html += '</ul></p>'
    html = '<div class="htmlText">%s</div>' % html
    return html
    
dd.inject_field(pcsw.Person,
    'cbss_summary',
    dd.VirtualField(dd.HtmlBox(_("CBSS summary")),cbss_summary))
    

MODULE_LABEL = _("CBSS")

#~ settings.LINO.add_user_field('cbss_level',UserLevels.field(MODULE_LABEL))
#~ settings.LINO.add_user_group('cbss',MODULE_LABEL)
#~ settings.LINO.add_user_field('cbss_level',UserLevels.field(MODULE_LABEL))

dd.inject_field('countries.City',
    'inscode',
    models.CharField(
        max_length=5,
        verbose_name=_("INS code"),
        blank=True,
    ))
        
dd.inject_field('countries.Country',
    'inscode',
    models.CharField(
        max_length=3,
        verbose_name=_("INS code"),
        blank=True,
    ))
        


def setup_site_cache(self,force):
    """
    Called from :meth:`build_site_cache`. 
    First argument is the LINO instance."""
    
    import logging
    logger = logging.getLogger(__name__)
    
    environment = settings.LINO.cbss_environment
    if not environment:
        return # silently return
        
    if not environment in CBSS_ENVS:
        raise Exception("Invalid `cbss_environment` %r: must be empty or one of %s." % (
          environment,CBSS_ENVS))
    
    context = dict(cbss_environment=environment)
    def make_wsdl(template,parts):
        fn = os.path.join(settings.MEDIA_ROOT,*parts) 
        if not force and os.path.exists(fn):
            if os.stat(fn).st_mtime > self.mtime:
                logger.info("NOT generating %s because it is newer than the code.",fn)
                return
        s = file(os.path.join(os.path.dirname(__file__),'WSDL',template)).read()
        s = s % context
        makedirs_if_missing(os.path.dirname(fn))
        open(fn,'wt').write(s)
        logger.info("Generated %s for environment %r.",fn,environment)
        
    make_wsdl('RetrieveTIGroupsV3.wsdl',RetrieveTIGroupsRequest.wsdl_parts)
    make_wsdl('WebServiceConnector.wsdl',SSDNRequest.wsdl_parts)
    #~ make_wsdl('TestConnectionService.wsdl',TestConnectionRequest.wsdl_parts)
    
    # The following xsd files are needed, unmodified but in the same directory
    #~ for fn in 'RetrieveTIGroupsV3.xsd', 'rn25_Release201104.xsd', 'TestConnectionServiceV1.xsd':
    for fn in 'RetrieveTIGroupsV3.xsd', 'rn25_Release201104.xsd':
        src = os.path.join(os.path.dirname(__file__),'XSD',fn)
        target = os.path.join(settings.MEDIA_ROOT,'cache','wsdl',fn) 
        if not os.path.exists(target):
            shutil.copy(src,target)
            
    
def site_setup(self):
    """
    (Called during site setup.)
    
    Adds a new tab "CBSS" to the Detail of `contacts.Persons`.
    """
    self.modules.contacts.AllPersons.add_detail_tab('cbss',"""
    cbss_identify_person cbss_manage_access cbss_retrieve_ti_groups
    cbss_summary
    """,MODULE_LABEL,required=dict(user_groups='cbss')
    )
    #~ required_user_groups=['cbss']
    #~ cbss.IdentifyRequestsByPerson
    #~ self.modules.contacts.AllPersons.add_detail_panel('cbssrequests',"""
    #~ cbss_identify_person 
    #~ cbss_manage_access 
    #~ cbss_retrieve_ti_groups
    #~ """,_("CBSS Requests"))
    #~ self.modules.contacts.AllPersons.add_detail_tab('cbss',"cbssrequests",MODULE_LABEL,required_user_groups=['cbss'])
    #~ 
    
    self.modules.lino.SiteConfigs.add_detail_tab('cbss',"""
    cbss_org_unit sector ssdn_user_id ssdn_email
    cbss_http_username cbss_http_password
    """,MODULE_LABEL,required=dict(user_groups='cbss'))
    
    
    
def setup_main_menu(site,ui,user,m): pass
def setup_master_menu(site,ui,user,m): pass
  
#~ def setup_my_menu(site,ui,user,m): 
    #~ if user.profile.cbss_level < UserLevels.user: 
        #~ return
    #~ m  = m.add_menu("cbss",MODULE_LABEL)
    #~ m.add_action(MyIdentifyPersonRequests)
    #~ m.add_action(MyManageAccessRequests)
    #~ m.add_action(MyRetrieveTIGroupsRequests)
    
def setup_config_menu(site,ui,user,m):
    m  = m.add_menu("cbss",MODULE_LABEL)
    m.add_action(Sectors)
    m.add_action(Purposes)
  
def setup_explorer_menu(site,ui,user,m):
    if user.profile.cbss_level < UserLevels.manager: 
        return
    m  = m.add_menu("cbss",MODULE_LABEL)
    m.add_action(IdentifyPersonRequests)
    m.add_action(ManageAccessRequests)
    m.add_action(RetrieveTIGroupsRequests)
    
dd.add_user_group('cbss',MODULE_LABEL)




# -*- coding: UTF-8 -*-
## Copyright 2006-2012 Luc Saffre
## This file is part of the Lino project.
## Lino is free software; you can redistribute it and/or modify 
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 3 of the License, or
## (at your option) any later version.
## Lino is distributed in the hope that it will be useful, 
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
## GNU General Public License for more details.
## You should have received a copy of the GNU General Public License
## along with Lino; if not, see <http://www.gnu.org/licenses/>.

r"""
``memo`` is a simple universal markup parser that expands "commands" 
found in an input string to produce a resulting output string. 
Commands are in the form ``[KEYWORD ARGS]``. 
The caller defines itself all commands, there are no predefined commands.

Usage example
-------------

Instatiate a parser:

>>> p = Parser()

We declare a "command handler" function `url2html` and register it:

>>> def url2html(s):
...     print "[DEBUG] url2html() got %r" % s
...     if not s: return "XXX"
...     url,text = s.split(None,1)
...     return '<a href="%s">%s</a>' % (url,text)
>>> p.register_command('url',url2html)

The intended usage of our example handler is 
``[url URL TEXT]``, where URL is the URL 
to link to, and TEXT is the label of the link:

>>> print p.parse('This is a [url http://xyz.com test].')
[DEBUG] url2html() got 'http://xyz.com test'
This is a <a href="http://xyz.com">test</a>.


A command handler will be called with one parameter:
the portion of text between the KEYWORD and the 
closing square bracket. Not including the whitespace 
after the keyword.
It must return the text which is to replace 
the ``[KEYWORD ARGS]`` fragment.
It is responsible for parsing 
the text that it receives as parameter.
Our example implementation has a bug, 
it doesn't support the case of having only an URL without TEXT:

>>> print p.parse('This is a [url http://xyz.com].')
[DEBUG] url2html() got 'http://xyz.com'
This is a [ERROR need more than 1 value to unpack in '[url http://xyz.com]' at position 10-30].

Newlines preceded by a backslash will be removed 
before the command handler is called:

>>> print p.parse('''This is [url http://xy\
... z.com another test].''')
[DEBUG] url2html() got 'http://xyz.com another test'
This is <a href="http://xyz.com">another test</a>.

The whitespace between the KEYWORD and ARGS can be any 
whitespace, including newlines:

>>> print p.parse('''This is a [url 
... http://xyz.com test].''')
[DEBUG] url2html() got 'http://xyz.com test'
This is a <a href="http://xyz.com">test</a>.

The ARGS part is optional
(it's up to the command handler to react accordingly,
our handler function returns XXX in that case):

>>> print p.parse('''This is a [url] test.''')
[DEBUG] url2html() got ''
This is a XXX test.

The ARGS part may contain pairs of square brackets:

>>> print p.parse('''This is a [url 
... http://xyz.com test with [more] brackets].''')
[DEBUG] url2html() got 'http://xyz.com test with [more] brackets'
This is a <a href="http://xyz.com">test with [more] brackets</a>.


Fragments of text between brackets that do not match 
any registered command will be left unchanged:

>>> print p.parse('''This is a [1] test.''')
This is a [1] test.

>>> print p.parse('''This is a [foo bar] test.''')
This is a [foo bar] test.

>>> print p.parse('''Text with only [opening square bracket.''')
Text with only [opening square bracket.


Limits
------

A single closing square bracket as part of ARGS will not 
produce the desired result:

>>> print p.parse('''This is a [url 
... http://xyz.com The character "\]"].''')
[DEBUG] url2html() got 'http://xyz.com The character "\\'
This is a <a href="http://xyz.com">The character "\</a>"].

Built-in commands like `[if ...]`, `[endif ...]`, `[=<expression>]`,... 
would be nice.


"""

import re

REGEX = re.compile(r"\[(\w+)\s*((?:[^[\]]|\[.*?\])*?)\]")

class Parser(object):
  
    def __init__(self):
        self.commands = dict()

    def register_command(self,cmd,func):
        self.commands[cmd] = func

    def cmd_match(self,matchobj):
        cmd = matchobj.group(1)
        params = matchobj.group(2)
        params = params.replace('\\\n','')
        h = self.commands.get(cmd,None)
        if h is None:
            return matchobj.group(0)
        try:
            return h(params)
        #~ except KeyError,e:
        except Exception,e:
            return self.handle_error(matchobj,e)
            
    def handle_error(self,mo,e):
        #~ return mo.group(0)
        return "[ERROR %s in %r at position %d-%d]" %(
          e, mo.group(0), mo.start(),mo.end())
        

    def parse(self,s):
        s = REGEX.sub(self.cmd_match,s)
        return s


def _test():
    import doctest
    doctest.testmod()

if __name__ == "__main__":
    _test()

